from typing import List, Dict, Any

from .checkers import check_type_filter


def format_filter_value(filter_type: str, filter_value: str) -> str:
    """Takes in a filter type and the value to filter for. Depenining on the type
    the value is formatted with the correct wild card e.g. 'contains' will add a
    wildcard to the start and end of value *value*

    Args:
        type (str): Either 'equal to', 'contains', 'starts with'
        value (str): value to filter for

    Raises:
        OAUnsupportedFilter: Raised when unsupported filter is requested such as 'ends with'

    Returns: Formatted filter with the value and correct filter wildcards.
    """

    filter_type = filter_type.lower()
    check_type_filter(filter_type, filter_value)

    if filter_type == "equal to":
        return f"{filter_value}"
    elif filter_type == "starts with":
        return f"{filter_value}*"
    elif filter_type == "contains":
        return f"*{filter_value}*"


def format_int_list_of_numbers(counrty_code: int, numbers: List[int]) -> List[str]:
    """Takes a list of integer numbers with no country code and the country code needed.
    This will then return a list of strings with the country code inserted infront of all
    numbers in orginal list.

    Args:
        counrty_code (int): Country code to be added infront of all numbers in list
        numbers (list): List of integer numbers with no country code

    Returns: List of numbers in string format with the country code added.
    """
    return [f"{counrty_code}-{number}" for number in sorted(numbers)]


def format_service_instance_profile(data: Dict) -> Dict[str, Any]:
    """Adds a blank dict if serviceInstanceProfile is not in data but needed

    Args:
        data (dict): Data to check if serviceInstanceProfile is in

    Returns: Empty Dict with serviceInstanceProfile key
    """

    data.setdefault("serviceInstanceProfile", {})
    return data


def sanitise_data(data: Dict):
    """Cleans data of any sensitive information that should not be leaked.

    Args:
        data (Dict): Request or response dict from API call
    """

    # data that hits this should only be of type dict
    if not isinstance(data, dict):
        return f"Unsupported data type {type(data)}"

    sanitised_data = data.copy()

    # removing sensitive data from log
    sensitive_keys = [
        "token",
        "access_token",
        "refresh_token",
        "api_key",
        "password",
    ]

    for key in sensitive_keys:
        if key in sanitised_data:
            sanitised_data[key] = (
                f"{sanitised_data[key][:2]}...{sanitised_data[key][-2:]}"
            )

    return sanitised_data



from .base_endpoint import BaseEndpoint


class CallCenters(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_group_call_centers(self, service_provider_id: str, group_id: str):
        """Retrieves a list of active call centers within a specified group, along with their settings.

        Args:
            service_provider_id (str): Target Service Provider where group is hosted
            group_id (str): Target Group ID

        Returns:
            List: List of Call Centers and their settings.
        """

        endpoint = "/groups/call-centers"
        params = {"serviceProviderId": service_provider_id, "groupId": group_id}
        return self._requester.get(endpoint, params=params)

    def get_group_call_center(self, service_user_id: str):
        """Retrieves deatiled information on a single Call Center.

        Args:
            service_user_id (str): Target Call Center's ID

        Returns:
            Dict: Target Call Centers details.
        """

        endpoint = "/groups/call-centers"

        params = {"serviceUserId": service_user_id}

        return self._requester.get(endpoint, params=params)

    def get_user_call_center(self, user_id: str):
        """Retrieves a list of call centers that the specified user is currently associated with.

        Args:
            user_id (str): Target User ID

        Returns:
            Dict: Agents Call Centers setting and a list of the User's associated Call Centers.
        """

        endpoint = "/users/call-center"

        params = {"userId": user_id}

        return self._requester.get(endpoint, params=params)

    def get_group_call_center_agents(self, service_user_id: str):
        """Returns list of agents assigned to the target call center.

        Args:
            service_user_id (str): Service user ID of target call center.

        Returns:
            Dict: List of agents assigned to call center
        """

        endpoint = "/groups/call-centers/agents"

        params = {"serviceUserId": service_user_id}

        return self._requester.get(endpoint, params=params)

    def get_group_call_center_bounced_calls(self, service_user_id: str):
        """Retrieves the number of rings before a call is bounced from the specified call center.

        Args:
            service_user_id (str): Target Call Center ID

        Returns:
            Dict: Amount of Rings before a call is Bounced
        """

        endpoint = "/groups/call-centers/bounced-calls"

        params = {"serviceUserId": service_user_id}

        return self._requester.get(endpoint, params=params)

    def get_group_call_center_forced_forwarding(self, service_user_id: str):
        """Retrieves the forwarding number for a call center if it is set to forward calls, along with any associated audio messages.

        Args:
            service_user_id (str): Target Call Center ID

        Returns:
            Dict: Number to be Forwarded to, alongside any Audio Messages.
        """

        endpoint = "/groups/call-centers/forced-forwarding"

        params = {"serviceUserId": service_user_id}

        return self._requester.get(endpoint, params=params)

    def get_group_call_center_overflow(self, service_user_id):
        """Retrieves the forwarding number for a user when all call center agents are busy, along with any associated audio messages.

        Args:
            service_user_id (str): Target Call Center ID

        Returns:
            Dict: Dict: Call Centers overflow configuration.
        """

        endpoint = "/groups/call-centers/overflow"

        params = {"serviceUserId": service_user_id}

        return self._requester.get(endpoint, params=params)

    def get_group_call_center_stranded_calls(self, service_user_id):
        """Retrieves the forwarding number for a user when a call center doesn't answer, along with any associated audio messages.

        Args:
            service_user_id (str): Target Call Center ID

        Returns:
            Dict: Call Centers stranded call configuration.
        """

        endpoint = "/groups/call-centers/stranded-calls"

        params = {"serviceUserId": service_user_id}

        return self._requester.get(endpoint, params=params)

    def get_group_call_center_stranded_calls_unavailable(self, service_user_id):
        """Retrieves the forwarding number for a user when a call center doesn't answer, along with the count of agents with an unavailable code in the call center.

        Args:
            service_user_id (str): Target Call Center ID

        Returns:
            Dict: Number to be Forwarded to, and Agents with an Unavailable Code set.
        """

        endpoint = "/groups/call-centers/stranded-calls-unavailable"

        params = {"serviceUserId": service_user_id}

        return self._requester.get(endpoint, params=params)

    # POST

    # PUT
    def put_group_call_centers_status(
        self, call_center_user_ids: list, status: bool = True
    ):
        """Updates a list of call centers (CC) status to either active or inactive.

        Args:
            call_center_user_ids(list): List of service user IDs (CC IDs), the status given will be applied to these.
            status (bool): Boolean value of True (Activate) or False (Deactivate) which will be applied to list of AAs.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/call-centers/status"

        data = {
            "instances": [
                {"serviceUserId": call_center_user_id, "isActive": status}
                for call_center_user_id in call_center_user_ids
            ]
        }

        return self._requester.put(endpoint, data=data)

    def put_group_call_center(self, call_center_user_id: str, updates: dict):
        """Allows you to update a specific call center.

        Args:
            call_center_user_id (str): Service user id of the target call center.
            updates (dict): Updates to apply in in a dictionary format.

        Returns:
            Dict: The call center with the new applied updates.
        """

        endpoint = "/groups/call-centers"

        updates["serviceUserId"] = call_center_user_id

        return self._requester.put(endpoint, data=updates)

    def put_group_call_center_agents(
        self, call_center_user_id: str, agent_user_ids: list
    ):
        """Add or remove agents to a call center.

        Note: Leave the agent_user_ids blank to remove all users and to remove some only include
        the users you would like to include in this call center.

        Args:
            call_center_user_id (str): Service user ID of the target call center.
            agent_user_ids (list): List of user IDs to be added to call center.

        Returns:
            Dict: Dictionary of the new state of the CC.
        """

        endpoint = "/groups/call-centers/agents"

        data = {
            "serviceUserId": call_center_user_id,
            "agents": [{"userId": agent_id} for agent_id in agent_user_ids],
        }

        return self._requester.put(endpoint, data=data)

    def put_group_call_center_agents_levels(
        self, call_center_user_id: str, agent_user_ids: list, skill_level: int
    ):
        """Update a list of agents skill level in a single Call Center (CC).

        Args:
            call_center_user_id (str): Service user ID of the call center users belong to.
            agent_user_ids (list): List of the target users.
            skill_level (int): Skill level that will be applied to the list of users in the target call center.

        Returns:
            Dict: CC ID and list of the agent and their updated skill level.
        """

        endpoint = "/groups/call-centers/agents"

        data = {
            "serviceUserId": call_center_user_id,
            "agents": [
                {"userId": agent_id, "skillLevel": skill_level}
                for agent_id in agent_user_ids
            ],
        }

        return self._requester.put(endpoint, data=data)

    def put_group_call_center_bounced_calls(
        self, call_center_user_id: str, updates: dict
    ):
        """Update the bounced call settings of a single Call Center (CC)

        Args:
            call_center_user_id (str): Service user ID of the target CC.
            updates (dict): Updates to apply to target CC.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/call-centers/bounced-calls"

        updates["serviceUserId"] = call_center_user_id

        return self._requester.put(endpoint, data=updates)

    def put_group_call_center_dnis_instance(
        self, call_center_user_id: str, updates: dict
    ):
        """Update a DNIS instance of a single Call Center (CC).

        Args:
            call_center_user_id (str): Service user ID of the target CC.
            updates (dict): Updates to apply to target CC.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/call-centers/dnis/instances"

        updates["serviceUserID"] = call_center_user_id

        return self._requester.put(endpoint, data=updates)

    def put_group_call_center_forced_forwarding(
        self, call_center_user_id: str, updates: dict
    ):
        """Update the forced forwarding settings of a single Call Center (CC).

        Args:
            call_center_user_id (str): Service user ID of the target CC.
            updates (dict): Updates to apply to target CC.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/call-centers/forced-forwarding"

        updates["serviceUserID"] = call_center_user_id

        return self._requester.put(endpoint, data=updates)

    def put_group_call_center_overflow(self, call_center_user_id: str, updates: dict):
        """Update the overflow settings of a single Call Center (CC).

        Args:
            call_center_user_id (str): Service user ID of the target CC.
            updates (dict): Updates to apply to target CC.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/call-centers/overflow"

        updates["serviceUserID"] = call_center_user_id

        return self._requester.put(endpoint, data=updates)

    def put_group_call_center_stranded_calls(
        self, call_center_user_id: str, updates: dict
    ):
        """Update the stranded calls settings of a single Call Center (CC).

        Args:
            call_center_user_id (str): Service user ID of the target CC.
            updates (dict): Updates to apply to target CC.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/call-centers/stranded-calls"

        updates["serviceUserID"] = call_center_user_id

        return self._requester.put(endpoint, data=updates)

    def put_group_call_center_stranded_calls_unavailable(
        self, call_center_user_id: str, updates: dict
    ):
        """Update the stranded calls unavailable settings of a single Call Center (CC).

        Args:
            call_center_user_id (str): Service user ID of the target CC.
            updates (dict): Updates to apply to target CC.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/call-centers/stranded-calls-unavailable"

        updates["serviceUserID"] = call_center_user_id

        return self._requester.put(endpoint, data=updates)

    def put_user_call_center_supervised_agents(
        self, call_center_user_id: str, supervisor_user_id: str, agent_ids: list
    ):
        """Update the list of agents a supervisor has in a single Call Center (CC).

        Args:
            call_center_user_id (str): Service user ID of the target CC.
            supervisor_user_id (str): User ID of the supervisor.
            agent_ids (list): List of user IDs of agents to apply to supervisor.

        Returns:
            Dict: Superivor ID and list of agents they supervise.
        """

        endpoint = "/groups/call-centers/supervisors"

        data = {
            "serviceUserId": call_center_user_id,
            "supervisorUserId": supervisor_user_id,
            "supervisors": [{"userId": agent_id} for agent_id in agent_ids],
        }

        return self._requester.put(endpoint, data=data)

    def put_user_call_center(self, user_id: str, updates: dict):
        """Update an agent's status in a Call Center (CC).

        Args:
            user_id (str): User ID of the target user.
            updates (dict): Updates to be applied to the user.

        Returns:
            Dict: Agents ACD status and status in each CC they are assigned to.
        """

        endpoint = "/users/call-center"

        updates["userId"] = user_id

        return self._requester.put(endpoint, data=updates)

    def put_user_call_center_agents_update(
        self, user_id: str, call_center_service_ids: list
    ):
        """Update the Call Centers (CC) a user is assigned to.

        Args:
            user_id (str): User ID of the target user.
            call_center_service_ids (list): List of CC service user IDs to update the user with.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/user/call-centers/agents"

        data = {
            "agentUserId": user_id,
            "callCenters": [
                {"userId": call_center_id} for call_center_id in call_center_service_ids
            ],
        }

        return self._requester.put(endpoint, data=data)

    def put_user_call_center_agent_sign_out(self, user_id: str):
        """Sign the user out of their assigned Call Centers (CC).

        Args:
            user_id (str): User ID of the target user.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/user/call-centers/agents/sign-out"

        data = {
            "agentUserId": user_id,
        }

        return self._requester.put(endpoint, data=data)


# DELETE



from .base_endpoint import BaseEndpoint


class Session(BaseEndpoint):
    def __init__(self):
        super().__init__()

    # GET

    def get_session(self):
        """Fetches session information for currently logged in user"""

        endpoint = "/auth/session"

        return self._requester.get(endpoint)

    # POST

    def post_session(self, username: str, password: str):
        """Requests access token.

        Args:
            username (str): Odin instance username
            password (str): Odin instance password
        """

        endpoint = "/auth/token"

        payload = {"username": username, "password": password}

        return self._requester.post(endpoint, data=payload)

    def post_session_switch(self, username: str):
        """Switch users

        Args:
            username (str): User you would like to switch to
        """

        endpoint = "/auth/switch-user"

        payload = {"username": username}

        return self._requester.post(endpoint, data=payload)

    def post_session_logout(self):
        """Logs out of session"""

        endpoint = "/auth/token/logout"

        payload = {"token": self._requester.token}

        return self._requester(endpoint, data=payload)

    # PUT

    def put_session(self):
        """Refreshes access token fetching a new one which will expire in 24hr"""

        endpoint = "/auth/token"

        return self._requester.put(endpoint)

    def put_change_password(self, user_id: str, new_password: str):
        """Change the password of a user

        Args:
            user_id (str): Tagrte user id of user to reset password
            new_password (str): New password
        """

        endpoint = "/users/passwords"

        updates = {"userId": user_id, "password": new_password}

        return self._requester.put(endpoint, data=updates)

    def put_password(self, user_id: str, old_password: str, new_password: str):
        """Change password of a specific user.

        Args:
            user_id (str): Target user to change password
            old_password (str): Current password user logs in with
            new_password (str): Password to change to
        """

        endpoint = "/auth/passwords"

        updates = {
            "userId": user_id,
            "oldPassword": old_password,
            "newPassword": new_password,
        }

        return self._requester.put(endpoint, data=updates)

    def put_my_password(self, old_password: str, new_password: str):
        """Change the password of the current user

        Args:
            old_password (str): Current password user logs in with
            new_password (str): Password to change to
        """

        endpoint = "/auth/password"

        updates = {"oldPassword": old_password, "newPassword": new_password}

        self._requester.put(endpoint, data=updates)

    # DELETE



from .base_endpoint import BaseEndpoint


class Authentication(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    # POST

    # PUT

    def put_user_authentication_service(self, user_id: str, new_password: str):
        """Set new SIP Authentication password for a single user.

        Args:
            user_id (str): Target user ID to reset the SIP authentication password.
            new_password (str): New SIP authentication password to apply to new user.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/users/authentication"

        data = {"userId": user_id, "newPassword": new_password}

        return self._requester.put(endpoint, data=data)

    def put_user_web_authentication_password(self, user_id: str, new_password: str):
        """Set new Web Authentication password for a single user.

        Args:
            user_id (str): Target user ID to reset the web authentication password.
            new_password (str): New web authentication password to apply to new user.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/users/passwords"

        data = {"userId": user_id, "newPassword": new_password}

        return self._requester.put(endpoint, data=data)

    # DELETE



def main(
    api,
    current_service_provider_id: str,
    current_group_id: str,
    start_of_range_number: str,
    end_of_range_number: str = None,
) -> bool:
    """Removes a singular or a range of numbers from a Broadworks instance."""

    logger = api.logger

    logger.info(f"Removing numbers from {current_group_id}")
    # delete number from group
    api.dns.delete_group_dns(
        current_service_provider_id,
        current_group_id,
        start_of_range_number=start_of_range_number,
        end_of_range_number=end_of_range_number,
    )

    logger.info(f"Removing numbers from {current_service_provider_id}")
    # remove from sp/ent
    api.dns.delete_service_provider_dns(
        current_service_provider_id,
        start_of_range_number=start_of_range_number,
        end_of_range_number=end_of_range_number,
    )

    return True



def main(api, service_provider_id: str, group_id: str, user_id: str):
    """identify a user's associations with Call Centers (CC), Hunt Groups (HG),
    and Pick Up Groups.

    Args:
        service_provider_id (str): Service Provider where the group is hosted.
        group_id (str): Group where the User is located.
        user_id (str): Target user ID.

    Returns:
        str: Formatted output of the user showing all CC, HG, and Pick Up user is assigned to.
    """

    logger = api.logger

    USER_DATA = {
        "userId": user_id,
        "firstName": None,
        "lastName": None,
        "extension": None,
        "phoneNumber": None,
        "aliases": None,
        "services": None,
        "featurePacks": None,
        "huntGroups": [],
        "callCenters": [],
        "pickUpGroup": None,
    }

    # fetch user, returns error if not found
    logger.info(f"Fetching user {user_id}")
    user = api.reports.get_user_report(user_id)

    USER_DATA["firstName"] = user["firstName"]
    USER_DATA["lastName"] = user["lastName"]
    USER_DATA["extension"] = user["extension"]
    USER_DATA["phoneNumber"] = user["phoneNumber"]
    USER_DATA["services"] = user["userServices"]
    USER_DATA["featurePacks"] = user["servicePacks"]
    USER_DATA["aliases"] = user["aliases"]

    logger.info("Fetching users pick up group")
    pick_up_group = api.call_pickup.get_call_pickup_group_user(
        service_provider_id, group_id, user_id
    )

    try:
        USER_DATA["pickUpGroup"] = pick_up_group[0]["name"]
    except IndexError:
        logger.error("User has no pickup group")
        USER_DATA["pickUpGroup"] = None

    logger.info("Fetching hunt groups")
    hunt_groups = api.hunt_groups.get_group_hunt_group_user(
        service_provider_id, group_id, user_id
    )
    for hg in hunt_groups:
        USER_DATA["huntGroups"].append(hg["serviceUserId"])

    # if the user does not have a license for CC this call errors
    try:
        logger.info("Fetching users call centers")
        call_centers = api.call_centers.get_user_call_center(user_id)
        for cc in call_centers["callCenters"]:
            USER_DATA["callCenters"].append(cc["serviceUserId"])
    except Exception:
        logger.error("user is not assigned to any call centers")
        USER_DATA["callCenters"] = None

    return USER_DATA



def main(api, service_provider_id: str, group_id: str):
    logger = api.logger

    # Dictionary Descripting Total Users Devices
    registrations_out = {}

    logger.info(f"Fetching groups {group_id} registrations")
    group_registration = api.regsitration.get_bulk_user_registration(
        service_provider_id, group_id
    )

    users = group_registration.get("users", [])

    logger.info("Analysing user registrations")
    for user in users:
        user_id = user["profile"]["userId"]
        registrations_out[user_id] = {"registration": {}}  # Initialise the dictionary

        for registration in user["data"]["registrations"]:
            device_name = registration["deviceName"]

            registrations_out[user_id]["registration"][device_name] = {
                "registered": True
            }  # All devices listed under the "registrations" attribute must be registered, so always True.

    return registrations_out



from .base_endpoint import BaseEndpoint
from ..exceptions import OSFileNotFound
import base64


class Announcements(BaseEndpoint):
    def __init__(self):
        super().__init__()

    # GET
    def get_user_announcements(self, user_id: str):
        """Retrieves announcements from the given user, including Anouncement Type and file size limits.

        Args:
            user_id (str): User ID of the target user you would like to retrieve the announcements from.

        Returns:
            Dict: Contains Announcement Info, and a list of Announcements.
        """

        endpoint = "/users/announcements"

        params = {"userId": user_id}

        return self._requester.get(endpoint, params=params)

    def get_group_announcements(self, group_id: str, service_provider_id: str):
        """Retrieves announcements from the given Group, including Anouncement Type and file size limits.

        Args:
            group_id (str): Group ID of the target Group you would like to retrieve the announcements from.
            service_provider_id (str): Service Provider ID of where the group resides

        Returns:
            Dict: Contains Announcement Info, and a list of Announcements.
        """

        endpoint = "/groups/announcements"

        params = {"groupId": group_id, "serviceProviderId": service_provider_id}

        return self._requester.get(endpoint, params=params)

    # POST

    def post_user_announcement(
        self,
        user_id: str,
        name: str,
        description: str,
        file_path: str,
    ):
        """Adds an announcement to the given user. Must be a WAV File!

        Args:
            user_id (str): User ID of the target user you would like to add the Announcement to.
            name (str): Desired name of the Announcement
            description (str): Description assigned to the Announcement.
            file_path (str): File path to the Announcement Audio File. (Raw String)
            media_type (str): Only supported type is WAV
        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/users/announcements"

        try:
            with open(
                file_path, "rb"
            ) as audio_file:  # Converts input audio file into b64
                content = base64.b64encode(audio_file.read()).decode("ascii")
        except FileNotFoundError:
            raise OSFileNotFound

        payload = {
            "userId": user_id,
            "name": name,
            "mediaType": "WAV",
            "description": description,
            "content": content,
        }

        return self._requester.post(endpoint, data=payload)

    def post_group_announcement(
        self,
        group_id: str,
        service_provider_id: str,
        name: str,
        description: str,
        file_path: str,
    ):
        """Adds an announcement to the given user. Must be a WAV file!

        Args:
            group_id (str): Group ID of the target Group you would like to add the Announcement to.
            service_provider_id (str): Service Provider ID of where the group resides
            name (str): Desired name of the Announcement
            description (str): Description assigned to the Announcement.
            file_path (str): File path to the Announcement Audio File. (Raw String)
            media_type (str): Only supported type is WAV

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/announcements"

        try:
            with open(
                file_path, "rb"
            ) as audio_file:  # Converts input audio file into b64
                content = base64.b64encode(audio_file.read()).decode("ascii")
        except FileNotFoundError:
            raise OSFileNotFound

        payload = {
            "groupId": group_id,
            "serviceProviderId": service_provider_id,
            "name": name,
            "mediaType": "WAV",
            "description": description,
            "content": content,
        }

        return self._requester.post(endpoint, data=payload)

    # PUT

    def put_user_announcement(self, user_id: str, name: str, new_name: str):
        """Updates the name of the given user Announcement

        Args:
            user_id (str): User ID of the user the announcement resides
            name (str): Current name of the Announcement
            media_type (str): Only supported type is WAV
            new_name (str): Desired name of the Announcement


        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/users/announcements"

        updates = {
            "userId": user_id,
            "name": name,
            "mediaType": "WAV",
            "newName": new_name,
        }

        return self._requester.put(endpoint, data=updates)

    def put_group_announcement(
        self,
        group_id: str,
        service_provider_id: str,
        name: str,
        new_name: str,
    ):
        """Updates the name of the given Group Announcement

        Args:
            group_id (str): Group ID of the target Group you would like to change the Announcement name of.
            service_provider_id (str): Service Provider ID of where the group resides
            name (str): Current name of the Announcement
            new_name (str): Desired name of the Announcement
            media_type (str): Only supported type is WAV


        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/announcements"

        updates = {
            "groupId": group_id,
            "serviceProviderId": service_provider_id,
            "name": name,
            "mediaType": "WAV",
            "newName": new_name,
        }

        return self._requester.put(endpoint, data=updates)

    # DELETE

    def delete_user_announcement(self, user_id: str, name: str):
        """Removes the given user Announcement

        Args:
            user_id (str): User ID of the user the announcement resides
            name (str): Current name of the Announcement
            media_type (str): Only supported type is WAV


        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/users/announcements"

        params = {
            "userId": user_id,
            "name": name,
            "mediaType": "WAV",
        }

        return self._requester.delete(endpoint, params=params)

    def delete_group_announcement(
        self,
        group_id: str,
        service_provider_id: str,
        name: str,
    ):
        """Removes the given Group Announcement

        Args:
            group_id (str): Group ID of the target Group you would like to remove
            service_provider_id (str): Service Provider ID of where the group resides
            media_type (str): Only supported type is WAV


        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/announcements"

        params = {
            "groupId": group_id,
            "serviceProviderId": service_provider_id,
            "name": name,
            "mediaType": "WAV",
        }

        return self._requester.delete(endpoint, params=params)



from .base_endpoint import BaseEndpoint
from ..utils.formatters import format_filter_value


class Users(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET
    def get_users(
        self,
        service_provider_id: str = None,
        group_id: str = None,
        filter: str = None,
        filter_type: str = None,
        filter_value: str = None,
        limit: int = None,
        extended=False,
    ):
        """
        Returns list of users depending on filter criteria.

        Args:
            servive_provider_id (str, optional): Service or Enterprise ID,
            top level object. Defaults to None.
            group_id (str, optional): Group ID where user is hosted. Defaults to None.
            filter (str, optional): Filter criteria, supported filters below. Defaults to None.
            filter_type (str, optional): Options: equal to, starts with, or contains. Defaults to None.
            filter_value (str, optional): Value filtering on e.g. firstName. Defaults to None.
            limit (int, optional): Limits the amount of values API returns. Defaults to None.

        Returns:
            dict: List of users.

        #Supported Filters
        macAddress: search by device
        lastName: filter by lastName
        firstName: filter by firstName
        dn: filter by dn
        emailAddress: filter by emailAddress
        userId: filter by userId
        extension: filter by extension

        ####Examples
        Get all users in Enterprise ent1
        GET /api/v2/users?serviceProviderId=ent1

        Get all users in Group grp1
        GET /api/v2/users?serviceProviderId=ent1&groupId=grp1

        Get up to 10 users in the system with a last name that contains smith
        GET /api/v2/users?lastName=*smith*&limit=10

        Get the users in grp1 that have a phone number that starts with 513333
        GET /api/v2/users?serviceProviderId=ent1&groupId=grp1&dn=513333*
        """

        endpoint = "/users?"

        params = {}

        if service_provider_id:
            params["serviceProviderId"] = service_provider_id
            if group_id:
                params["groupId"] = group_id
        if filter:
            params[filter] = format_filter_value(filter_type, filter_value)
        if limit:
            params["limit"] = limit
        if extended:
            params["extended"] = True

        return self._requester.get(endpoint, params=params)

    def get_user_by_id(self, user_id: str):
        """Returns extensive details of a single user including alias, enpoint device, and more common
        details like first and last name.

        Args:
            user_id (str): Target user ID of the user you would like to review.

        Returns:
            Dict: Python dictionary of the users details
        """

        endpoint = "/users"

        params = {"userId": user_id}

        return self._requester.get(endpoint, params=params)

    # POST
    def post_user(
        self,
        service_provider_id: str,
        group_id: str,
        user_id: str,
        first_name: str,
        last_name: str,
        extension: str,
        web_auth_password: str,
        payload: dict = {},
    ):
        """
            Creates a new user in the specified group with the configuration defined in the payload.

        Args:
            service_provider_id (str): Service provider ID where Group is loctaed.
            group_id (str): Group ID where new user will be built.
            user_id (str): Complete User ID including group domain of new user.
            first_name (str): First name of new user.
            last_name (str): Last name of new user.
            extension (str): Extension number of new user.
            web_auth_password (str): Web authentication password. Note get.password_generate() can be used to get this.
            payload (dict, optional): User configuration.

        Returns:
            Dict: New user entity.
        """

        endpoint = "/users"

        payload["callingLineIdFirstName"] = (
            first_name
            if not payload.get("callingLineIdFirstName")
            else payload["callingLineIdFirstName"]
        )
        payload["callingLineIdLastName"] = (
            last_name
            if not payload.get("callingLineIdLastName")
            else payload["callingLineIdLastName"]
        )

        payload["serviceProviderId"] = service_provider_id
        payload["groupId"] = group_id
        payload["userId"] = user_id
        payload["firstName"] = first_name
        payload["lastName"] = last_name
        payload["extension"] = extension
        payload["password"] = web_auth_password

        return self._requester.post(endpoint, data=payload)

    # PUT
    def put_users_bulk(self, users: list, updates: dict):
        """
        Updates specified list of User's options, such as extension, name and etc.

        Note: Available options to change can be seen through: get.user_by_id()

        Args:
            users (list): List of specified User IDs to update
            updates (dict): The updates to be applied to the list of Users e.g {"extension":"9999"}

        Returns:
            Dict: Returns the changes made including the list of User IDs and updates.
        """
        endpoint = "/users/bulk"

        target_users = [{"userId": user} for user in users]

        data = {"users": target_users, "data": updates}

        return self._requester.put(endpoint, data=data)

    def put_user(
        self, service_provider_id: str, group_id: str, user_id: str, updates: dict
    ):
        """
        Updates specified User's options, such as extension, name and etc.

        Note: Available options to change can be seen through: get.user_by_id()

        Args:
            service_provider_id (str):
            updates (dict): The updates to be applied Target Service Provider where group is located
            group_id (str): Target Group ID where user is located
            user_id (str): Target User IDto the list of Users e.g {"extension":"9999"}

        Returns:
            Dict: Returns the changes made including User ID and updates.
        """
        endpoint = "/users"

        updates["serviceProviderId"] = service_provider_id
        updates["groupId"] = group_id
        updates["userId"] = user_id

        return self._requester.put(endpoint, data=updates)

    def put_user_portal_passcode(self, user_id: str, new_passcode: str):
        """Updates the specified User's portal passcode.

        Args:
            user_id (str): User ID of the target user you would like to change the portal passcode for.
            new_passcode (int): New portal passcode to set for the target user.

        Raises:
            AOInvalidCode: If code is less than 4 or higher than 6.

        Returns:
            None: This method does not return any specific value.
        """

        if len(new_passcode) < 4 or len(new_passcode) > 6:
            from ..exceptions import OSInvalidCode

            raise OSInvalidCode

        endpoint = "/users/portal-passcode"

        data = {"userId": user_id, "newPasscode": new_passcode}

        return self._requester.put(endpoint, data=data)

    # DELETE



from .base_endpoint import BaseEndpoint


class Devices(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    # POST
    def post_group_device(
        self,
        service_provider_id: str,
        group_id: str,
        device_name: str,
        device_type: str,
        payload: dict = {},
    ):
        """Adds a new device to a group.

        Args:
            service_provider_id (str): Service provider ID where the device should be built.
            group_id (str): Group ID where the device should be built
            device_name (str): Name of the new device
            device_type (str): Type of device.
            payload (dict, optional): Device configuration data.

        Returns:
            Dict: Returns the device profile.
        """

        endpoint = "/groups/devices"

        payload["serviceProviderId"] = service_provider_id
        payload["groupId"] = group_id
        payload["deviceName"] = device_name
        payload["deviceType"] = device_type

        return self._requester.post(endpoint, data=payload)

    # PUT
    def put_group_devices(
        self, service_provider_id: str, group_id: str, device_name: str, updates: dict
    ):
        """Update a single device in a group.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where Group is located.
            group_id (str): Group ID where target device is located.
            device_name (str): Device name of the target device.
            updates (dict): Updates to apply to the target device.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """

        endpoint = "/groups/devices"

        updates["serviceProviderId"] = service_provider_id
        updates["groupId"] = group_id
        updates["deviceName"] = device_name

        return self._requester.put(endpoint, data=updates)

    def put_service_provider_device(
        self, service_provider_id: str, device_name: str, updates: dict
    ):
        """Update a single device in a Service Provider or Enterprise.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where device is located.
            device_name (str): Device name of the target device.
            updates (dict): Updates to apply to the target device.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """

        endpoint = "/service-providers/devices"

        updates["serviceProviderId"] = service_provider_id
        updates["deviceName"] = device_name

        return self._requester.put(endpoint, data=updates)

    def put_system_devices(self, device_name: str, updates: dict):
        """Update a single device in the Broadworks system.

        Args:
            device_name (str): Device name of the target device.
            updates (dict): Updates to apply to the target device.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """

        endpoint = "/service-providers/devices"

        updates["deviceName"] = device_name

        return self._requester.put(endpoint, data=updates)

    def put_system_device_file(self, device_name: str, updates: dict):
        """Update a config file for a single device at the system level.

        Args:
            device_name (str): Device name of the target device.
            updates (dict): Updates to apply to the target device.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/system/devices/files"

        updates["deviceName"] = device_name

        return self._requester.put(endpoint, data=updates)

    def put_service_provider_device_file(self, device_name: str, updates: dict):
        """Update a config file for a single device at Service Provider or
        Enterprise level.

        Args:
            device_name (str): Device name of the target device.
            updates (dict): Updates to apply to the target device.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/service_provider/devices/files"

        updates["deviceName"] = device_name

        return self._requester.put(endpoint, data=updates)

    def put_group_device_file(self, device_name: str, updates: dict):
        """Update a config file for a single device at Group level.

        Args:
            device_name (str): Device name of the target device.
            updates (dict): Updates to apply to the target device.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/devices/files"

        updates["deviceName"] = device_name

        return self._requester.put(endpoint, data=updates)

    def put_group_device_tags_profile(
        self, service_provider_id: str, group_id: str, device_name: str, tags: list
    ):
        """Update tags assigned to single device at group level.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where Group is located.
            group_id (str): Group ID where target device is located.
            device_name (str): Device name of the target device.
            tags (dict): List of dictionaries tag name and value entries.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """

        endpoint = "/groups/devices/profile"

        data = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "deviceName": device_name,
            "tags": [{"elements": tags}],
        }

        return self._requester.put(endpoint, data=data)

    def put_group_device_tag(
        self,
        service_provider_id: str,
        group_id: str,
        device_name: str,
        tag_name: str,
        tag_value: str,
    ):
        """Update a single tag assigned to a device at group level.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where Group is located.
            group_id (str): Group ID where target device is located.
            device_name (str): Device name of the target device.
            tag_name (str): Name of the tag to add or update.
            tag_value (str): Value of tag to add or update.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """

        endpoint = "/groups/devices/tags"

        data = {
            "tagName": f"%{tag_name}%",
            "tagValue": tag_value,
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "deviceName": device_name,
        }

        return self._requester.put(endpoint, data=data)

    def put_service_provider_device_tag(
        self, service_provider_id: str, device_name: str, tag_name: str, tag_value: str
    ):
        """Update a single tag assigned to a device at the Service Provider or Enterprise level.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where Group is located.
            device_name (str): Device name of the target device.
            tag_name (str): Name of the tag to add or update.
            tag_value (str): Value of tag to add or update.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """

        endpoint = "/service-providers/devices/tags"

        data = {
            "tagName": f"%{tag_name}%",
            "tagValue": tag_value,
            "serviceProviderId": service_provider_id,
            "deviceName": device_name,
        }

        return self._requester.put(endpoint, data=data)

    def put_system_device_tag(self, device_name: str, tag_name: str, tag_value: str):
        """Update a single tag assigned to a device at the System level.

        Args:
            device_name (str): Device name of the target device.
            tag_name (str): Name of the tag to add or update.
            tag_value (str): Value of tag to add or update.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """

        endpoint = "/system/devices/tags"

        data = {
            "tagName": f"%{tag_name}%",
            "tagValue": tag_value,
            "deviceName": device_name,
        }

        return self._requester.put(endpoint, data=data)

    def put_group_device_type_file(
        self, service_provider_id: str, group_id: str, device_type: str, updates: dict
    ):
        """Set config file for all devices of a specific type at the group level.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where Group is located.
            group_id (str): Group ID where target device is located.
            device_type (str): The device type you would like to apply the changes to.
            updates (dict): Cupdates (dict): Updates to apply to the target device.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/device-types/files"

        updates["serviceProviderId"] = service_provider_id
        updates["groupId"] = group_id
        updates["deviceType"] = device_type

        return self._requester.put(endpoint, data=updates)

    def put_group_device_type_tag(
        self,
        service_provider_id: str,
        group_id: str,
        device_type: str,
        tag_name: str,
        tag_value: str,
    ):
        """Update tags applied to device types at the Group level.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where Group is located.
            group_id (str): Group ID where target device is located.
            device_type (str): The target device type to apply the updates.
            tag_name (str): Name of the tag to add or update.
            tag_value (str): Value of tag to add or update.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """

        endpoint = "/groups/system/device-types/tags"

        data = {
            "tagName": tag_name,
            "tagValue": tag_value,
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "deviceType": device_type,
        }

        return self._requester.put(endpoint, data=data)

    def put_service_provider_device_type_tag(
        self, service_provider_id: str, device_type: str, tag_name: str, tag_value: str
    ):
        """Update tags applied to device types at the Service Provider or Enterprise level.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where Group is located.
            device_type (str): The target device type to apply the updates.
            tag_name (str): Name of the tag to add or update.
            tag_value (str): Value of tag to add or update.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """

        endpoint = "/service-providers/device-types/tags"

        data = {
            "tagName": tag_name,
            "tagValue": tag_value,
            "serviceProviderId": service_provider_id,
            "deviceType": device_type,
        }

        return self._requester.put(endpoint, data=data)


# DELETE



def main(
    api,
    service_provider_id,
    group_id,
    user_id,
    device_type,
    email,
    primary_device,
    webex_feature_pack_name,
    enable_integarated_imp,
):
    logger = api.logger

    # Update the email & alt user id
    email_alt_userid = {
        "emailAddress": email,
        "alternateUserId": [{"alternateUserId": email, "description": "Webex"}],
    }

    try:
        api.users.put_user(
            service_provider_id, group_id, user_id, updates=email_alt_userid
        )
        logger.info("Updated email and alt user ID.")
    except Exception as e:
        logger.error(f"Failed to update users email and alt user ID, detail: {e}")

    # Assign feature pack
    try:
        if webex_feature_pack_name:
            api.services.put_user_services(
                user_id=user_id, service_packs=[webex_feature_pack_name]
            )
        logger.info(f"Added feature {webex_feature_pack_name}")
    except Exception as e:
        logger.error(f"Failed to add feature {webex_feature_pack_name}, detail: {e}")

    # enable IMP in service settings
    if enable_integarated_imp:
        enable_IMP = {"Integrated IMP": {"isActive": True}}
        try:
            api.services.put_user_service_settings(user_id=user_id, settings=enable_IMP)
            logger.info("Enabled integrated IMP")
        except Exception as e:
            logger.info(f"Failed to enable Integrated IMP, detail: {e}")

    # build device
    device_name = f"{user_id.split('@')[0]}_WBX"
    device_password = api.password_generate.get_password_generate(
        service_provider_id, group_id
    )["password"]

    device_payload = {
        "useCustomUserNamePassword": "true",
        "accessDeviceCredentials": {
            "userName": device_name,
            "password": device_password,
        },
        "userName": device_name,
    }

    try:
        api.devices.post_group_device(
            service_provider_id=service_provider_id,
            group_id=group_id,
            device_name=device_name,
            device_type=device_type,
            payload=device_payload,
        )
        logger.info("Built device")
    except Exception as e:
        logger.error(f"Failed to build device, detail: {e}")

    # Add device to user based on primary flag - JP
    if primary_device:
        primary_device_configuration = {
            "endpointType": "accessDeviceEndpoint",
            "accessDeviceEndpoint": {
                "accessDevice": {
                    "deviceType": device_type,
                    "deviceName": device_name,
                    "serviceProviderId": service_provider_id,
                    "groupId": group_id,
                    "deviceLevel": "Group",
                },
                "linePort": user_id,
            },
        }
        try:
            api.users.put_user(
                service_provider_id, group_id, user_id, primary_device_configuration
            )
            logger.info("Added device to user as primary")
        except Exception as e:
            logger.error(f"failed to add device as primary, detail: {e}")
    else:
        try:
            api.shared_call_appearance.post_user_shared_call_appearance_endpoint(
                user_id, user_id.replace("@", "_WBX@"), device_name
            )
            logger.info("Added device as shared call appearance")
        except Exception as e:
            logger.error(f"Failed to add devie as shared call appearance, detail: {e}")

    # Get webex password
    try:
        password = api.password_generate.get_password_generate(
            service_provider_id, group_id
        )["password"]
        api.authentication.put_user_web_authentication_password(user_id, password)
        logger.info("Set webex password")
    except Exception as e:
        logger.error(f"Failed to set webex password, Detail {e}")

    # Return formatted data
    webex_user_details = {
        "username": email,
        "password": password,
        "primary_device": primary_device,
        "device_type": device_type,
    }

    return webex_user_details



def main(api, service_provider_id: str, group_id: str):
    """Audits a group for chargeable services this will return all feature packs
    assigned and count for all broadwork entities such as users, call centers, hunt groups etc.
    Additionaly this will return all DID/DDI's and their active status.

    :param service_provider_id: Service provider or Enterprise ID that contains group being audited.
    :param group_id: Group that is to be audited.

    :return r:
    """

    # save logger from api
    logger = api.logger

    # all Services
    logger.info("Fetching all group services")
    service_report = api.services.get_group_services(group_id, service_provider_id)

    assigned_user_services = []
    assigned_group_services = []
    assigned_service_pack_services = []

    logger.info("Analysing user services")
    for us in service_report["userServices"]:
        if us["usage"] > 0:
            del us["authorized"]
            del us["assigned"]
            del us["limited"]
            del us["quantity"]
            del us["licensed"]
            del us["allowed"]
            del us["userAssignable"]
            del us["groupServiceAssignable"]
            del us["tags"]
            del us["alias"]

            logger.info(f"Fetching users assigned {us}")
            users = api.services.get_group_services_user_assigned(
                group_id, service_provider_id, us["serviceName"], "serviceName"
            )
            userIDs = [u["userId"] for u in users["users"]]
            us["users"] = userIDs

            assigned_user_services.append(us)

    logger.info("Analysing group services")
    for gs in service_report["groupServices"]:
        if gs["usage"] > 0:
            del gs["authorized"]
            del gs["assigned"]
            del gs["limited"]
            del gs["quantity"]
            del gs["licensed"]
            del gs["allowed"]
            del gs["instanceCount"]
            del gs["alias"]
            assigned_group_services.append(gs)

    logger.info("Analysing service packs")
    for sps in service_report["servicePackServices"]:
        if sps["usage"] > 0:
            del sps["authorized"]
            del sps["assigned"]
            del sps["limited"]
            del sps["allowed"]
            del sps["serviceName"]
            del sps["quantity"]
            del sps["alias"]

            logger.info(f"Fetching users assigned {us}")
            users = api.services.get_group_services_user_assigned(
                group_id, service_provider_id, sps["servicePackName"], "servicePackName"
            )
            userIDs = [u["userId"] for u in users["users"]]
            sps["users"] = userIDs

            assigned_service_pack_services.append(sps)

    # Group DNs
    logger.info("Fetching group dns")
    dn_report = api.dns.get_group_dns(service_provider_id, group_id)
    all_dns = {
        "assigned": {"activated": [], "deactivated": []},
        "unassigned": {"activated": [], "deactivated": []},
    }

    logger.info("Analysing group dns")
    for dn in dn_report["dns"]:
        if dn["assigned"] and dn["activated"]:
            all_dns["assigned"]["activated"] += dn["list"]
        elif dn["assigned"] and not dn["activated"]:
            all_dns["assigned"]["deactivated"] += dn["list"]
        elif not dn["assigned"] and dn["activated"]:
            all_dns["unassigned"]["activated"] += dn["list"]
        elif not dn["assigned"] and not dn["activated"]:
            all_dns["unassigned"]["deactivated"] += dn["list"]

    total_assigned_activated = len(all_dns["assigned"]["activated"])
    total_assigned_deactivated = len(all_dns["assigned"]["deactivated"])
    total_unassigned_activated = len(all_dns["unassigned"]["activated"])
    total_unassigned_deactivated = len(all_dns["unassigned"]["deactivated"])
    total_dns = (
        total_assigned_activated
        + total_assigned_deactivated
        + total_unassigned_activated
        + total_unassigned_deactivated
    )

    all_dns["totalDNs"] = total_dns
    all_dns["assigned"]["totalAssignedDNs"] = (
        total_assigned_activated + total_assigned_deactivated
    )
    all_dns["unassigned"]["totalUassignedDNs"] = (
        total_unassigned_activated + total_unassigned_deactivated
    )

    # Group Detail
    group_detail = api.groups.get_group(service_provider_id, group_id)

    # Trunking detail
    logger.info("Fetching trunking capacity")
    try:
        trunk_detail = api.trunk_groups.get_group_trunk_groups_call_capacity(
            service_provider_id, group_id
        )
        del trunk_detail["serviceProviderId"]
        del trunk_detail["groupId"]
    except Exception:
        trunk_detail = []

    group_audit = {
        "groupDetail": group_detail,
        "licenceBreakdown": {
            "userServices": assigned_user_services,
            "groupServices": assigned_group_services,
            "servicePackServices": assigned_service_pack_services,
        },
        "groupDNs": all_dns,
        "groupTrunking": trunk_detail,
    }

    return group_audit



from .data_store import DataStore
from .broadwork_entities import (
    ServiceProvider,
    Group,
    TrunkGroup,
    AAKey,
    AAMenu,
    AutoAttendant,
    CallCenter,
    HuntGroup,
    User,
    Contact,
    Address,
    Department,
)

__all__ = [
    "DataStore",
    "ServiceProvider",
    "Group",
    "TrunkGroup",
    "AAKey",
    "AAMenu",
    "AutoAttendant",
    "CallCenter",
    "HuntGroup",
    "User",
    "Contact",
    "Address",
    "Department",
]



from .base_endpoint import BaseEndpoint


class TrunkGroups(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET
    def get_group_trunk_groups_call_capacity(
        self, service_provider_id: str, group_id: str
    ):
        """Fetches Trunk Call Capacity data for a single Group.

        Args:
            service_provider_id (str): Service Provider/ Enterprise ID where Group is located.
            group_id (str): Target Group to return data on.

        Returns:
            Dict: Trunk Group Call Capacity data of target group.
        """

        endpoint = "/groups/trunk-groups/call-capacity"

        params = {"groupId": group_id, "serviceProviderId": service_provider_id}

        return self._requester.get(endpoint, params=params)

    def get_group_trunk_group(
        self, service_provider_id: str, group_id: str, trunk_group_name: str
    ):
        """Fetches all Trunk Group details of a single Trunk Group in a Group.

        Args:
            service_provider_id (str): Service Provider/ Enterprise ID where Group is located.
            group_id (str): Group ID where the target Trunk Group is located.
            trunk_group_name (str): Target Trunk Group Name.

        Returns:
            Dict: Details of a target trunk group.
        """

        endpoint = "/groups/trunk-groups"

        params = {
            "groupId": group_id,
            "serviceProviderId": service_provider_id,
            "name": trunk_group_name,
        }

        return self._requester.get(endpoint, params=params)

    def get_group_trunk_groups(self, service_provider_id: str, group_id: str):
        """Fetches list of all trunk groups in a single group.

        Args:
            service_provider_id (str): Service Provider/ Enterprise ID where Group is located.
            group_id (str): Group ID where the target Trunk Group is located.

        Returns:
            List: List of core details of all Trunk Groups located in a single Group.
        """

        endpoint = "/groups/trunk-groups"

        params = {"groupId": group_id, "serviceProviderId": service_provider_id}

        return self._requester.get(endpoint, params=params)

    def get_service_provider_trunk_group_call_capacity(self, service_provider_id: str):
        """Fetches trunk call capacity details of a single Service Provider.

        Args:
            service_provider_id (str): Target Service Provider/ Enterprise ID.

        Returns:
            Dict: Trunk call capacity details of a single Service Provider/ Enterprise ID.
        """

        endpoint = "/service-providers/trunk-groups/call-capacity"

        params = {"serviceProviderId": service_provider_id}

        return self._requester.get(endpoint, params=params)

    def get_service_provider_trunk_call_capacity_report(self, service_provider_id: str):
        """Fetches trunk call capacity details of Service Provider/ Enterprise and all Groups in the SP/ ENT.

            Args:
                servive_provider_id (str): Target Service Provider/ Enterprise ID.

            Returns:
                Dict: Breakdown of all trunk call capacity details of target Service Provider/ Enterprise and all Groups \
                    in the target SP/ ENT.
            """

        endpoint = "/service-providers/trunk-groups/call-capacity/reports"

        params = {"serviceProviderId": service_provider_id}

        return self._requester.get(endpoint, params=params)

    # POST
    def post_group_trunk_group(
        self,
        service_provider_id: str,
        group_id: str,
        trunk_name: str,
        max_active_calls: int,
        payload: dict = {},
        sip_authentication_username: str = "",
        sip_authentication_password: str = "",
    ):
        """
        Builds a Trunk Group (TG) in the specified group.

        Args:
            service_provider_id (str): The service provider ID in which the target group is built.
            group_id (str): The group ID where the HG should be built.
            trunk_name (str): The name of the new TG.
            max_active_calls (str): The maximum active calls to be set on the TG.
            payload (dict, optional): Configuration for the TG.
            sip_authentication_username (str, optional): The SIP authentication username for the TG. This field is required if "requireAuthentication" is set to "true".
            sip_authentication_password (str, optional): The SIP authentication password for the TG. You can generate a password for this using get.sip_password_generator. This field is required if "requireAuthentication" is set to "true".

        Note:
            Several fields are set to have default values. Please refer to the online documentation.

        Returns:
            Dict: Returns the Trunk Group profile.
        """

        endpoint = "/groups/trunk-groups"

        payload["name"] = trunk_name
        payload["maxActiveCalls"] = max_active_calls
        payload["serviceProviderId"] = service_provider_id
        payload["groupId"] = group_id

        default_payload_values = {
            "capacityExceededTrapInitialCalls": 0,
            "capacityExceededTrapOffsetCalls": 0,
            "clidSourceForScreenedCallsPolicy": "Profile Name Profile Number",
            "continuousOptionsSendingIntervalSeconds": 30,
            "failureOptionsSendingIntervalSeconds": 10,
            "failureThresholdCounter": 1,
            "invitationTimeout": 6,
            "inviteFailureThresholdCounter": 1,
            "inviteFailureThresholdWindowSeconds": 30,
            "pilotUserCallOptimizationPolicy": "Optimize For User Services",
            "pilotUserCallingLineAssertedIdentityPolicy": "Unscreened Originating Calls",
            "pilotUserCallingLineIdentityForEmergencyCallsPolicy": "No Calls",
            "pilotUserCallingLineIdentityForExternalCallsPolicy": "No Calls",
            "pilotUserChargeNumberPolicy": "No Calls",
            "successThresholdCounter": 1,
            "useSystemUserLookupPolicy": "true",
            "userLookupPolicy": "Basic",
            "requireAuthentication": "false",
        }
        for key, default_value in default_payload_values.items():
            payload.setdefault(key, default_value)

        if payload["requireAuthentication"]:
            payload["sipAuthenticationUserName"] = sip_authentication_username
            payload["sipAuthenticationPassword"] = sip_authentication_password

        return self._requester.post(endpoint, data=payload)

    # PUT

    def put_group_trunk_groups_call_capacity(
        self,
        service_provider_id: str,
        group_id: str,
        max_active_calls: int,
        bursting_max_active_calls: int = None,
    ):
        """Updates the trunking call capacity in the specified group.

        NOTE: The max available active calls cannot be changed here. Please see service_providers_trunk_group_call_capacity to update this.

        Args:
            service_provider_id (str): Service provider ID where the target group is built
            group_id (str): Group ID whose trunk group call capacity needs updating
            max_active_calls (int): The max active calls for the group
            bursting_max_active_calls (int, optional): The bursting max active calls for the group. Defaults to None.
        """

        endpoint = "/groups/trunk-groups/call-capacity"

        updates = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "maxActiveCalls": max_active_calls,
        }

        if bursting_max_active_calls is not None:
            updates["burstingMaxActiveCalls"] = bursting_max_active_calls

        return self._requester.put(endpoint, data=updates)

    def put_group_trunk_group(
        self,
        service_provider_id: str,
        group_id: str,
        trunk_group_name: str,
        updates: dict,
    ):
        """Updates trunk group (TG) configuration.

        Args:
            service_provider_id (str): Service provider ID where the target group is built.
            group_id (str): Group ID whose trunk group call capacity needs updating.
            trunk_group_name (str): The name of the trunk group that is being updated.
            updates (dict, optional): Updates to be applied to the TG.
        """

        endpoint = "/groups/trunk-groups"

        updates["serviceProviderId"] = service_provider_id
        updates["groupId"] = group_id
        updates["name"] = trunk_group_name

        return self._requester.put(endpoint, data=updates)

    def put_service_providers_trunk_group_call_capacity(
        self,
        service_provider_id: str,
        max_active_calls: int,
        bursting_max_active_calls: int = None,
    ):
        """Updates the max active calls and the bursting max active calls for the given service provider.

        Args:
            service_provider_id (str): Service provider ID for which the max active calls needs to be updated
            max_active_calls (int): The updates to be applied to the service provider's trunking call capacity
            bursting_max_active_calls (int, optional): Number of bursting calls. Defaults to none, do not include if only changing max active calls.
        """

        endpoint = "/service-providers/trunk-groups/call-capacity"

        updates = {
            "serviceProviderId": service_provider_id,
            "maxActiveCalls": max_active_calls,
        }

        if bursting_max_active_calls is not None:
            updates["burstingMaxActiveCalls"] = bursting_max_active_calls

        return self._requester.put(endpoint, data=updates)

    # DELETE



from ..exceptions import OSExtensionNotFound, OSRangeFault


def retrieve_extensions(api, service_provider_id: str, group_id: str, logger) -> list:
    extensions = []

    logger.info("Fetching users, hunt groups, call centers, and auto attendants")
    dataset = (
        api.users.get_users(service_provider_id, group_id)
        + api.hunt_groups.get_group_hunt_groups(service_provider_id, group_id)
        + api.call_centers.get_group_call_centers(service_provider_id, group_id)
        + api.auto_attendants.get_auto_attendants(service_provider_id, group_id)
    )

    logger.info("Extracting used extensions")
    for data in dataset:
        if not data["extension"]:
            continue

        extensions.append(int(data["extension"]))

    return extensions if extensions else None


def main(
    api, service_provider_id: str, group_id: str, range_start: int, range_end: int
):
    """Retrieves The Lowest Free Extension Available In The Designated Group Passed."""

    logger = api.logger

    # Raise Exception If Range Start Is Greater Than Range End
    if range_start > range_end:
        logger.error(f"Range_start {range_start} is larger than range_end {range_end}")
        raise OSRangeFault

    # Retrieve List Of Occupied Extensions Within The Group
    extensions = retrieve_extensions(api, service_provider_id, group_id, logger)

    logger.info("Filtering for useable extension")
    for extension in range(range_start, range_end + 1):
        if extension not in extensions:
            return {"extension": extension}

    logger.error(f"No useable extensions found in range {range_start} - {range_end}")
    raise OSExtensionNotFound



from .base_endpoint import BaseEndpoint


class PasswordGenerate(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_password_generate(self, service_provider_id: str, group_id: str) -> dict:
        """Generates a single passwords following the groups rules.

        Args:
            service_provider_id (str): Service Provider ID where Group is located.
            group_id (str): Group ID to generate password for.

        Returns:
            dict: Single password generated according to the groups rules.
        """

        endpoint = "/password/generate"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self._requester.get(endpoint, params=params)

    def get_passwords_generate(
        self, service_provider_id: str, group_id: str, limit: int = 10
    ) -> dict:
        """Generates a multiple passwords to the limit set in pararmeters.

        Args:
            service_provider_id (str): Service Provider ID where Group is located.
            group_id (str): Group ID to generate password for.
            limit (int, optional): Number of passwords api will return. Defaults to 10.

        Returns:
            dict: Multiple passwords generated according to the groups rules.
        """

        endpoint = "/password/generate"

        params = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "limit": limit,
        }

        return self._requester.get(endpoint, params=params)

    def get_passcode_generate(self, service_provider_id: str, group_id: str) -> dict:
        """Generates a single passcode following group rules.

        Args:
            service_provider_id (str): Service Provider ID where Group is located.
            group_id (str): Group ID to generate passcode for.

        Returns:
            dict: Single passcode generated according to the groups rules.
        """

        endpoint = "/passcode/generate"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self._requester.get(endpoint, params=params)

    def get_passcodes_generate(
        self, service_provider_id: str, group_id: str, limit: int = 10
    ) -> dict:
        """Generates a multiple passcodes to the limit set in pararmeters.

        Args:
            service_provider_id (str): Service Provider ID where Group is located.
            group_id (str): Group ID to generate passcodes for.
            limit (int, optional): Number of passwords api will return. Defaults to 10.

        Returns:
            dict: Multiple passcodes generated according to the groups rules.
        """

        endpoint = "/passcode/generate"

        params = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "limit": limit,
        }

        return self._requester.get(endpoint, params=params)

    def get_sip_password_generate(self) -> dict:
        """Generates a single SIP password.

        Args:
            service_provider_id (str): Service Provider ID where Group is located.
            group_id (str): Group ID to generate SIP password for.

        Returns:
            dict: Single SIP password generated according to the groups rules.
        """

        endpoint = "/sip-password/generate"

        return self._requester.get(endpoint)

    def get_sip_passwords_generate(self, limit: int = 10) -> dict:
        """Generates multiple SIP passwords to the limit set in parameters. Defaults to 10.

        Args:
            limit (int, optional): Number of passwords api will return. Defaults to 10.

        Returns:
            dict: Mutliple SIP passwords generated according to the groups rules.
        """

        endpoint = "/password/generate"

        params = {"limit": limit}

        return self._requester.get(endpoint, params=params)


# POST

# PUT

# DELETE



from .base_endpoint import BaseEndpoint


class Administrators(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    # POST

    def post_service_provider_admin(
        self,
        service_provider_id: str,
        user_id: str,
        password: str,
        first_name: str,
        last_name: str,
        language: str = "English",
        admin_type: str = "Normal",
        payload: dict = {},
    ):
        """Builds service provider administrator account.

        Args:
            service_provider_id (str): Target service parovider to build account.
            user_id (str): User ID of new account i.e. 'firstname.lastname@domain.com'
            password (str): Web access password user needs to log in.
            first_name (str): First name.
            last_name (str): Last name.
            language (str, optional): Find supported languages on your system. Defaults to "English".
            admin_type (str, optional): Type of admin, options: 'normal', 'customer', 'password reset only'. Defaults to normal.
            payload (dict, optional): Payload sent to API. Defaults to empty dict and is formatted for you.
        """

        endpoint = "/service-providers/admins"

        payload["serviceProviderId"] = service_provider_id
        payload["language"] = language
        payload["administratorType"] = admin_type.capitalize()
        payload["userId"] = user_id
        payload["password"] = password
        payload["firstName"] = first_name
        payload["lastName"] = last_name

        return self._requester.post(endpoint, data=payload)

    def post_group_admin(
        self,
        service_provider_id: str,
        group_id: str,
        user_id: str,
        password: str,
        payload: dict = {},
    ):
        """Builds a group-level administrator.

        Args:
            service_provider_id (str): Service provider ID where the admin should be built.
            group_id (str): Group ID where the admin should be built
            user_id (str): User ID of the admin.
            password (str): Password for the administrator profile. Note get.password_generate() can be used to get this.
            payload (dict, optional): Admin configuration data.

        Returns:
            Dict: Returns the admin profile.
        """

        endpoint = "/groups/admins"

        payload["serviceProviderId"] = service_provider_id
        payload["groupId"] = group_id
        payload["userId"] = user_id
        payload["password"] = password

        return self._requester.post(endpoint, data=payload)

    def post_group_admin_policies_bulk(self, user_ids: list, policy_config: dict):
        """Applies policy settings to multiple group administrators.

        Note: See docs for formatting of parameters.

        Args:
            user_ids (list): User IDs of admins to apply policy to.
            policy_config (dict): Policy settings to apply to target users.

        Returns:
            Dict: Returns admins and policy applied.
        """
        endpoint = "/groups/admins/policies/bulk"

        data = {"users": [{"userId": user} for user in user_ids], "data": policy_config}

        self._requester.post(endpoint, data=data)

    # PUT

    def put_service_provider_admin_policies(self, user_id: str, policy_config: dict):
        """AI is creating summary for put_service_provider_admin_policies

        Args:
            user_id (str): User ID of target service provider admin.
            policy_config (dict): Policy settings to update.
        """

        endpoint = "/service-providers/admins/policies"

        policy_config["userId"] = user_id

        return self._requester.put(endpoint, data=policy_config)

    # DELETE



from ..exceptions import OSUnsupportedFilter, OSUnsupportedFilterType
from .constants import supported_filters, supported_filter_types


def check_type_filter(filter_by: str, filter_type: str) -> None:
    """Checks if the filter and filter type is supported by the API.

    Args:
        filter_by (str): The data you would like to filter by e.g. firstName
        filter_type (str): The type of filter you would like to apply e.g. equal to

    Raises:
        OSUnsupportedFilter: Raised when the filter is not supported by the API
        OSUnsupportedFilterType: Raised when the filter type is not supported by the API

    Supported filters:
    - macAddress
    - lastName
    - firstName
    - dn
    - emailAddress
    - userId
    -extension

    Supported filter types:
    - equal to
    - starts with
    - contains
    """

    if filter_by not in supported_filters:
        raise OSUnsupportedFilter(filter_by)
    if filter_type not in supported_filter_types:
        raise OSUnsupportedFilterType(filter_type)



from .base_endpoint import BaseEndpoint

from ..utils.formatters import format_service_instance_profile as format_sip


class HuntGroups(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_group_hunt_groups(self, service_provider_id, group_id):
        """Returns a list of all the Hunt Groups within the specified Group.

        Args:
            service_provider_id (str): Target Service Provider ID
            group_id (str): Target Group ID

        Returns:
            List: Returns a list of every Hunt Group within a Group, alongside their extension and name.
        """

        endpoint = "/groups/hunt-groups"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self._requester.get(endpoint, params=params)

    def get_group_hunt_group(self, service_user_id):
        """Returns detailed information about the specified Hunt Group.
        Args:
            service_user_id (str): UserID of the target Hunt Group.

        Returns:
            Dict: Returns the specified Hunt Groups settings and information, such as group policies, agents, and extension.
        """

        endpoint = "/groups/hunt-groups"

        params = {"serviceUserId": service_user_id}

        return self._requester.get(endpoint, params=params)

    def get_group_hunt_group_user(self, service_provider_id, group_id, user_id):
        """Returns the Hunt Group's the specified User is apart of.

        Args:
            service_provider_id (str): Target Service Provider ID
            group_id (str): Target Group ID
            user_id (str): Target User ID

        Returns:
            List: Returns the Hunt Group's a user is within, alongside their settings within that Hunt Group.
        """

        endpoint = "/groups/hunt-groups/user"

        params = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "userId": user_id,
        }

        return self._requester.get(endpoint, params=params)

    def get_group_hunt_groups_available_users(
        self, service_provider_id: str, group_id: str
    ):
        """Returns a list of all users within the service provider that are available to be assigned to a hunt group in the specified group.

        Args:
            service_provider_id (str): Target Service Provider ID
            group_id (str): Target Group ID

        Returns:
            List: available users (dict)
        """

        endpoint = "/groups/hunt-groups/users"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self._requester.get(endpoint, params=params)

    # POST

    def post_group_hunt_group(
        self,
        service_provider_id: str,
        group_id: str,
        service_user_id: str,
        clid_first_name: str,
        clid_last_name: str,
        extension: str,
        payload: dict = {},
        agents: list = [],
        policy: str = "Regular",
        no_answer_number_of_rings: int = 5,
        forward_timeout_seconds: int = 0,
    ):
        """
        Builds a hunt group (HG) in the specified group.

        Args:
            service_provider_id (str): The service provider ID in which the target group is built.
            group_id (str): The group ID where the HG should be built.
            service_user_id (str): The service user ID for the new HG. This must include the domain of the user.
            clid_first_name (str): The Calling Line ID first name.
            clid_last_name (str): The Calling Line ID last name.
            extension (str): The extension number for the HG. This must be entered as a string.
            payload (dict, optional): HG configuration data.
            agents (list, optional): List of user IDs (str) that should be assigned to the new HG. The user(s) must already exist in the group.
            policy (str, optional): Regular, Circular, Simultaneous, Uniform, Weighted. Defaults to Regular.
            no_answer_number_of_rings (int, optional): Defaults to 5.
            forward_timeout_seconds (int, optional): Defaults to 0.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/hunt-groups"

        payload["serviceProviderId"] = service_provider_id
        payload["groupId"] = group_id
        payload["serviceUserId"] = service_user_id
        payload["policy"] = policy
        payload["noAnswerNumberOfRings"] = no_answer_number_of_rings
        payload["forwardTimeoutSeconds"] = forward_timeout_seconds

        if agents:
            payload["agents"] = [{"userId": agent} for agent in agents]

        payload = format_sip(payload)

        payload["serviceInstanceProfile"]["callingLineIdFirstName"] = clid_first_name
        payload["serviceInstanceProfile"]["callingLineIdLastName"] = clid_last_name
        payload["serviceInstanceProfile"]["name"] = (
            f"{clid_first_name} {clid_last_name}"
        )
        payload["serviceInstanceProfile"]["extension"] = extension

        return self._requester.post(endpoint, data=payload)

    def post_group_hunt_groups_remove_user(
        self, service_provider_id: str, group_id: str, user_id: str
    ):
        """Removes the specified user from all hunt groups in which it currently exists.

        Args:
            service_provider_id (str): The service provider ID in which the target user exists.
            group_id (str): The group ID where the user exists.
            user_id (str): The User ID of the user that is to be removed from the hunt group(s).

        Returns:
            List: Service user ID's (str) of the hunt groups from which the specified user has been removed.
        """

        endpoint = "/groups/hunt-groups/removeUser"

        data = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "userId": user_id,
        }

        return self._requester.post(endpoint, data=data)

    # PUT

    def put_group_hunt_groups_status(self, service_user_ids: list, status: bool = True):
        """Updates a list of Hunt Groups (HG) status to either active or inactive.

        Args:
            service_user_ids (list): List of service user IDs of target HGs.
            status (bool, optional): Status to apply to target HGs. Defaults to True.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/hunt-groups/status"

        data = {
            "instances": [
                {"serviceUserId": service_user_id, "isActive": status}
                for service_user_id in service_user_ids
            ]
        }

        return self._requester.put(endpoint, data=data)

    def put_group_hunt_group(
        self,
        service_provider_id: str,
        group_id: str,
        service_user_id: str,
        updates: dict,
    ):
        """Update a Hunt Group's (HG) settings.

        Args:
            service_provider_id (str): Service provider ID of where the group that hosts the HG is located.
            group_id (str): Group ID of where the HG is located.
            service_user_id (str): Target service user ID of the HG.
            updates (dict): Updates to be applied to HG.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/hunt-groups"

        updates = format_sip(updates)

        updates["serviceProviderId"] = service_provider_id
        updates["groupId"] = group_id
        updates["serviceUserId"] = service_user_id

        return self._requester.put(endpoint, data=updates)

    def put_group_hunt_group_weighted_call_distribution(
        self, service_provider_id: str, group_id, service_user_id: str, agents: list
    ):
        """Update the Weighted Call Distribution (WCD) between users in a Hunt Group (HG).

        Args:
            service_provider_id (str): Service provider ID where the group is located.
            group_id (_type_): Group ID where the HG is located.
            service_user_id (str): Service user ID of the target HG.
            agents (list): Updates of WCD to be applied to HG.

        Raises:
            AOInvalidWeighting: The WCD must equal 100 if it does not this error will be returned.

        Returns:
            None: This method does not return any specific value.
        """
        from ..exceptions import OSInvalidWeighting

        endpoint = "/groups/hunt-groups/weighted-call-distribution"

        data = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "serviceUserId": service_user_id,
            "agents": agents,
        }

        max_weights = 100
        assigned_weight = 0

        for agent in agents:
            assigned_weight += agent["weight"]
        if not assigned_weight == max_weights:
            raise OSInvalidWeighting

        return self._requester.put(endpoint, data=data)

    # DELETE

    def delete_group_hunt_group(self, service_user_id: str):
        """Deletes the specified hunt group.

        Args:
            service_user_id (str): The service user ID of the hunt group to be deleted.

        Returns:
            Dict: Profile of the deleted hunt group.

        """

        endpoint = "/groups/hunt-groups"

        params = {"serviceUserId": service_user_id}

        return self._requester.delete(endpoint, params=params)



from .base_endpoint import BaseEndpoint


class Groups(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_groups(self, service_provider_id: str) -> list:
        """Returns the specificied Service Provider's Groups.

        Args:
            service_provider_id (str): Target Service Provider ID

        Returns:
            List: List of groups and their Names, alongside groupID's and userLimits.
        """

        endpoint = "/groups"

        params = {"serviceProviderId": service_provider_id}

        return self._requester.get(endpoint, params=params)

    def get_group(self, service_provider_id: str, group_id: str) -> dict:
        """Returns the specificied Group's settings and information.

        Args:
            service_provider_id (str): Target Service Provider ID
            group_id (str): Target Group ID

        Returns:
            Dict: Returns information about the specified group, such as the DID, userCount and Domain.
        """

        endpoint = "/groups"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self._requester.get(endpoint, params=params)


# POST

# PUT

# DELETE



import os
import json

import shutil

from ..exceptions import OSFileNotFound


def check_directory_or_file_exists(directory_file_path: str) -> bool:
    """Checks if a directory or file exists.

    Args:
        directory_file_path (str): Path to the directory or file.

    Returns:
        bool: If path exists return True else False.
    """

    if os.path.exists(directory_file_path):
        return True

    return False


def join_path(directory: str, file_name: str) -> str:
    """Using os.path this method joins directory with file name and normalises path
    for the OS running on.

    Args:
        directory (str): Directory path.
        file_name (str): Name of file.

    Returns:
        str: Returns normalised string of joined path.
    """

    return os.path.normpath(os.path.join(directory, file_name))


def json_fie_to_dict(file_path: str) -> dict:
    """Loads a json file into code as Python dict.

    Args:
        file_path (str): Path to json file including file name.

    Returns:
        dict: Python dict of json file. Returns False if file not found.

    Raises:
        OSFileNotFound: Raised when file cant found.
    """

    if check_directory_or_file_exists(file_path):
        with open(file_path, "r") as data:
            return json.loads(data.read())

    return OSFileNotFound


def make_directory(directory_path: str) -> None:
    """Checks if directory already exists if not it will create it.

    Args:
        directory_path (str): Path to target directory.

    Returns:
        None: Function builds directory.
    """

    return os.makedirs(directory_path, exist_ok=True)


def copy_all_directorys_files_to_target(source_dir, target_dir) -> bool:
    """

    Args:
        source_dir (_type_): _description_
        target_dir (_type_): _description_

    Returns:
        bool: Returns True if operation succeeded.

    Raises:
        OSFileNotFound: Raised when source directoty can't be found.
    """

    if check_directory_or_file_exists(source_dir):
        # Check if target dir exists if not build it
        make_directory(target_dir)

        # Copy files
        shutil.copytree(source_dir, target_dir)
        return True

    return OSFileNotFound


def copy_single_file_to_target_directory(
    source_dir: str, target_dir: str, file_name: str
) -> bool:
    """Copies a single targeted file from a source directory to a target directory.

    Args:
        source_dir (str): Source directory path where target file is located.
        target_dir (str): Target directory path where target file is to be copied to.
        file_name (str): Name of source file

    Returns:
        bool: Returns True if operation succeeded.

    Raises:
        OSFileNotFound: Raised when source target file can't be found.
    """

    # Construct the full paths using os.path.join and normalize them
    source_file = os.path.normpath(os.path.join(source_dir, file_name))
    target_file = os.path.normpath(os.path.join(target_dir, file_name))

    if check_directory_or_file_exists(source_file):
        # Create the target directory if it does not exist
        os.makedirs(target_dir, exist_ok=True)

        # Copy the file to the target directory
        shutil.copy2(source_file, target_file)
        return True

    return OSFileNotFound


def remove_directory(directory_path: str) -> bool:
    """Check if file exists and type is directory and if both are true the directory is removed.

    Args:
        directory_path (str): Path to target directory to be removed.

    Returns:
        bool: _description_
    """

    if check_directory_or_file_exists(directory_path) and os.path.isdir(directory_path):
        shutil.rmtree(directory_path)
        return True

    return False



import graphviz

from ...store import broadwork_entities as bre


class GraphvizModule:
    """
    Colours:
    - Black: 020300
    - Orange: FFA400 - Darker: B87700
    - Red: 95190C - Darker: 5E1008
    - Purple: 5B5F97 - Darker: 3D4066
    - Redish: A23E48 - Darker: 68272E
    """

    NODE_STYLING = {
        "start": {
            "shape": "record",
            "style": "filled",
            "margin": "0.2",
            "color": "#020300",
            "fontname": "Arial",
            "fontcolor": "white",
        },
        "exit": {
            "shape": "record",
            "style": "filled",
            "margin": "0.2",
            "color": "#020300",
            "fontname": "Arial",
            "fontcolor": "white",
        },
        "auto_attendant": {
            "shape": "record",
            "style": "filled",
            "margin": "0.2",
            "color": "#1C3A3A",
            "fillcolor": "#356969",
            "fontname": "Arial",
            "fontcolor": "white",
        },
        "call_centre": {
            "shape": "record",
            "style": "filled",
            "margin": "0.2",
            "color": "#CCCCCC",
            "fillcolor": "#D9D9D9",
            "fontname": "Arial",
            "fontcolor": "black",
        },
        "hunt_group": {
            "shape": "record",
            "style": "filled",
            "margin": "0.2",
            "color": "#4773EB",
            "fillcolor": "#5E85ED",
            "fontname": "Arial",
            "fontcolor": "black",
        },
        "user": {
            "shape": "record",
            "style": "filled",
            "margin": "0.2",
            "color": "#CAE188",
            "fillcolor": "#E3FE97",
            "fontname": "Arial",
            "fontcolor": "black",
        },
    }
    EDGE_STYLYING = {"fontname": "Arial"}

    def __init__(self, output_directory: str = None):
        self.dot = graphviz.Digraph()
        self.output_directory = output_directory

    def generate_call_flow_graph(self, nodes: list, number: str):
        self.dot.attr(
            name=f"calls_to_{number}",
            label=number,
            fontname="Arial",
            fontcolor="white",
            rankdir="LR",
        )

        # build nodes
        self.dot.node("Start", "Start", GraphvizModule.NODE_STYLING["start"])
        for n in nodes:
            try:
                node_config = f"<extension> Extension: {n.extension} "
            except AttributeError:
                node_config = ""

            if isinstance(n, bre.User):
                node_config += f"| <name> Name: {n.first_name} {n.last_name}"
                self.dot.node(n.id, node_config, GraphvizModule.NODE_STYLING["user"])

            elif isinstance(n, bre.CallCenter) or isinstance(n, bre.HuntGroup):
                node_config += f"| <name> Name: {n.name} | <policy> Policy: {n.policy}"
                for i, a in enumerate(n.agents):
                    node_config += f"| <{a.id}> Agent {i+1}: {a.extension}"
                self.dot.node(
                    n.service_user_id,
                    node_config,
                    GraphvizModule.NODE_STYLING["call_centre"]
                    if isinstance(n, bre.CallCenter)
                    else GraphvizModule.NODE_STYLING["hunt_group"],
                )

            elif isinstance(n, bre.AutoAttendant):
                node_config += f"| <name> Name: {n.name}"
                for k in n.business_hours_menu.keys:
                    node_config += f"| <key{k.number}> Option {k.number}: {k.action}"
                    k.id = f"{n.service_user_id}:<key{k.number}>"
                self.dot.node(
                    n.service_user_id,
                    node_config,
                    GraphvizModule.NODE_STYLING["auto_attendant"],
                )

        # build edges
        for n in nodes:
            try:
                if n._start_node:
                    try:
                        self.dot.edge("Start", n.id)
                    except Exception:
                        self.dot.edge("Start", n.service_user_id)
            except AttributeError:
                # node is note the start node
                pass

            if isinstance(n, bre.User):
                if n.call_forwarding_always:
                    self._format_edge(n, n.call_forwarding_always, "CFA")
                if n.call_forwarding_busy:
                    self._format_edge(n, n.call_forwarding_busy, "CFB")
                if n.call_forwarding_not_reachable:
                    self._format_edge(n, n.call_forwarding_not_reachable, "CFNR")

            elif isinstance(n, bre.CallCenter):
                if n.bounced_calls_enabled:
                    self._format_edge(
                        n, n.bounced_calls_transfer_to_phone_number, "BCT"
                    )
                if n.overflow_calls_action:
                    self._format_edge(
                        n, n.overflow_calls_transfer_to_phone_number, "OF"
                    )
                if n.stranded_calls_action:
                    self._format_edge(
                        n, n.stranded_calls_transfer_to_phone_number, "SCF"
                    )
                if n.stranded_call_unavailable_action:
                    self._format_edge(
                        n, n.stranded_call_unavailable_transfer_to_phone_number, "USCF"
                    )

            elif isinstance(n, bre.HuntGroup):
                if n.forward_after_timeout_enabled:
                    self._format_edge(n, n.no_answer_forward_to_phone_number, "NACF")
                if n.call_forward_not_reachable_enabled:
                    self._format_edge(
                        n, n.call_forward_not_reachable_transfer_to_phone_number, "CFNR"
                    )

            elif isinstance(n, bre.AutoAttendant):
                for key in n.business_hours_menu.keys:
                    if "Transfer" in key.action:
                        self._format_edge(key, key.phone_number, key.number)

    def _format_edge(self, node_a: str, node_b: str, label: str):
        if node_b is None:
            return None

        node_a_id = node_a.id if hasattr(node_a, "id") else node_a.service_user_id
        node_b_id = node_b.id if hasattr(node_b, "id") else node_b.service_user_id

        self.dot.edge(node_a_id, node_b_id, label, GraphvizModule.EDGE_STYLYING)

    def _save_graph(self, filename: str):
        self.dot.render(
            directory=self.output_directory,
            filename=filename,
            format="svg",
            cleanup=True,
        ).replace("\\", "/")



from dataclasses import dataclass, field
from typing import List


@dataclass
class call_flow:
    name: str
    nodes: List = field(default_factory=list)


@dataclass
class external_number:
    id: str


@dataclass
class call_records_statistics:
    first_name: str
    last_name: str
    extension: str
    userId: str
    feature_packs: list
    total: int
    totalAnsweredAndMissed: str
    answeredTotal: str
    missedTotal: str
    busyTotal: str
    redirectTotal: str
    receivedTotal: str
    receivedMissed: str
    receivedAnswered: str
    placedTotal: str
    placedMissed: str
    placedAnswered: str

    @classmethod
    def from_dict(cls, first_name, last_name, extension, data):
        return cls(
            first_name=first_name,
            last_name=last_name,
            extension=extension,
            userId=data.get("userId"),
            feature_packs=data.get("servicePackServices"),
            total=data.get("total"),
            totalAnsweredAndMissed=str(data.get("totalAnsweredAndMissed")),
            answeredTotal=data.get("answeredTotal"),
            missedTotal=data.get("missedTotal"),
            busyTotal=data.get("busyTotal"),
            redirectTotal=data.get("redirectTotal"),
            receivedTotal=data.get("receivedTotal"),
            receivedMissed=data.get("receivedMissed"),
            receivedAnswered=data.get("receivedAnswered"),
            placedTotal=data.get("placedTotal"),
            placedMissed=data.get("placedMissed"),
            placedAnswered=data.get("placedAnswered"),
        )



from ..exceptions import OSInvalidPasswordType


def main(
    api, service_provider_id: str, group_id: str, users: list, password_type: str
) -> list:
    # save logger from api
    logger = api.logger

    # SIP auth password
    if password_type.lower() == "sip":
        logger.info("Generating new SIP passwords")
        new_passwords = api.password_generate.get_sip_passwords_generate(len(users))[
            "passwords"
        ]
        users_and_new_passwords = list(zip(users, new_passwords))

        logger.info("Setting new SIP passwords")
        for user in users_and_new_passwords:
            api.authentication.put_user_authentication_service(user[0], user[1])

        logger.info("Setting new SIP passwords complete")
        return [
            {"userId": user[0], "newPassword": user[1]}
            for user in users_and_new_passwords
        ]

    # intigration password
    elif password_type.lower() == "web":
        logger.info("Generating new SIP passwords")
        new_passwords = api.password_generate.get_passwords_generate(
            service_provider_id, group_id, len(users)
        )["passwords"]
        users_and_new_passwords = list(zip(users, new_passwords))

        logger.info("Setting new SIP passwords")
        for user in users_and_new_passwords:
            api.authentication.put_user_web_authentication_password(user[0], user[1])

        logger.info("Setting new SIP passwords complete")
        return [
            {"userId": user[0], "newPassword": user[1]}
            for user in users_and_new_passwords
        ]

    # voicemail/ portal
    elif password_type.lower() == "vm":
        logger.info("Generating new voicemail passcodes")
        new_passcodes = api.password_generate.get_passcodes_generate(
            service_provider_id, group_id, len(users)
        )["passcodes"]
        users_and_new_passcodes = list(zip(users, new_passcodes))

        logger.info("Setting new voicemail passcodes")
        for user in users_and_new_passcodes:
            api.users.put_user_portal_passcode(user[0], user[1])

        logger.info("Setting new voicemail passcodes complete")
        return [
            {"userId": user[0], "newPasscode": user[1]}
            for user in users_and_new_passcodes
        ]

    else:
        # raise error if user attempts to change another type of password.
        logger.error(f"Password type {password_type} not supported")
        raise OSInvalidPasswordType



from .base_endpoint import BaseEndpoint

from ..utils.formatters import format_service_instance_profile as format_sip


class AutoAttendants(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_auto_attendants(self, service_provider_id: str, group_id: str):
        """Returns a complete list of all Auto Attendants in a single group.

        Args:
            service_provider_id (str): Service Provider where Group is hosted.
            group_id (str): Target Group where Auto Attendants are hosted.

        Returns:
            List: List of Auto Attendants with basic info on them.
        """

        endpoint = "/groups/auto-attendants"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self._requester.get(endpoint, params=params)

    def get_auto_attendant(self, service_user_id: str):
        """Returns detailed information of a single Auto Attendant.

        Args:
            service_user_id (str): User ID of target Auto Attendant.

        Returns:
            dict: Detailed information of target Auto Attendant.
        """

        endpoint = "/groups/auto-attendants"

        params = {"serviceUserId": service_user_id}

        return self._requester.get(endpoint, params=params)

    def get_auto_attendant_user(
        self, service_provider_id: str, group_id: str, user_id: str
    ):
        """Returns detailed information about all Auto Attendants (AA) built in the same group as the specified user.

        Args:
            service_provider_id (str): Service Provider ID of the group where the user is built.
            group_id (str): Group ID of the group where the user is built.
            user_id (str): User ID of the user being queried.

        Returns:
           List: Returns a list of the AAs built in the group.
        """

        endpoint = "/groups/auto-attendants/user"

        params = {
            "userId": user_id,
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
        }

        return self._requester.get(endpoint, params=params)

    def get_auto_attendant_submenus(self, service_user_id: str):
        """Returns a list of the submenus of the specified Auto Attendant (AA). Works with Standard AAs only, basic AAs do not have submenus.

        Args:
            service_user_id (str): The service user ID of the AA

        Returns:
            List: Returns a list of the submenus associated with the AA.
        """

        endpoint = "/groups/auto-attendants/submenus"

        params = {"serviceUserId": service_user_id}

        return self._requester.get(endpoint, params=params)

    def get_auto_attendant_submenu_usage(self, service_user_id: str, submenu_id: str):
        """Returns the type of the specified Auto Attendant (AA) submenu. NOTE: This method does not return any usage data.

        Args:
            service_user_id (str): The service user ID of the AA being queried.
            submenu_id (str): The submenu ID of the submenu being queried.

        Returns:
            List: Returns a list containing a single dict of the submenu.
        """

        endpoint = "/groups/auto-attendants/submenus/usage"

        params = {"serviceUserId": service_user_id, "submenuId": submenu_id}

        return self._requester.get(endpoint, params=params)

    # POST

    def post_auto_attendant(
        self,
        service_provider_id: str,
        group_id: str,
        service_user_id: str,
        aa_name: str,
        aa_type: str = "Basic",
        payload: dict = {},
    ):
        """Builds an Auto Attendant (AA) from the given payload.

        Args:
            service_provider_id (str): Service Provider ID of the group where the AA should be built.
            group_id (str): Group ID where the AA should be built.
            service_user_id (str): Service User ID of the AA (including the domain).
            aa_name (str): Name of the AA
            aa_type (str, optional): Type of AA: "Basic" or "Standard". Will default to "Basic". NOTE: The "Auto Attendant - Standard" service must be enabled on the group in order for the aa_type to be set to "Standard".
            payload (dict, optional): Additional AA configuration data.

        Returns:
            Dict: Returns the AA profile.
        """

        endpoint = "/groups/auto-attendants"

        payload = format_sip(payload)

        payload["serviceProviderId"] = service_provider_id
        payload["groupId"] = group_id
        payload["serviceUserId"] = service_user_id
        payload["serviceInstanceProfile"]["name"] = aa_name
        payload["type"] = aa_type

        default_payload_values = {
            "enableVideo": "false",
            "extensionDialingScope": "Group",
            "nameDialingScope": "Group",
            "nameDialingEntries": "LastName + FirstName",
            "firstDigitTimeoutSeconds": 1,
        }

        for key, default_value in default_payload_values.items():
            payload.setdefault(key, default_value)

        return self._requester.post(endpoint, data=payload)

    def post_auto_attendant_remove_user(
        self, service_provider_id: str, group_id: str, user_id: str
    ):
        """Returns a list of the available Auto Attendants (AAs) built in the same group as the specified user.

        Args:
            service_provider_id (str): Service Provider ID where the user is built.
            group_id (str): Group ID where the user is built.
            user_id (str): User ID of the user.

        Returns:
            List: List of the Service User IDs of the AAs in the group.
        """

        endpoint = "/groups/auto-attendants/removeUser"

        payload = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "userId": user_id,
        }

        return self._requester.post(endpoint, data=payload)

    def post_auto_attendant_submenu(
        self,
        service_user_id: str,
        submenu_id: str,
        announcement_selection: str = "Default",
        extension_dialing: bool = True,
    ):
        """Posts a new submenu to the specified Auto Attendant (AA).

        Args:
            service_user_id (str): Service User ID of the AA.
            submenu_id (str): ID of the submenu to be created.
            announcement_selection (str, optional): "Default" or "Personal". Defaults to "Default".
            extension_dialing (bool, optional): Whether Level Extension Dialing is enabled or not. Defaults to True.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/auto-attendants/submenus"

        payload = {
            "serviceUserId": service_user_id,
            "submenuId": submenu_id,
            "announcementSelection": announcement_selection,
            "enableLevelExtensionDialing": extension_dialing,
        }

        return self._requester.post(endpoint, data=payload)

    # PUT

    def put_auto_attendants_status(
        self, auto_attendant_user_ids: list, status: bool = True
    ):
        """Updates a list of auto attendants (AA) status to either active or inactive.

        Args:
            auto_attendant_user_ids (list): List of service user IDs (AA IDs), the status given
            will be applied to these.
            status (bool, optional): Boolean value of True (Activate) or False (Deactivate)
            which will be applied to list of AAs. Defaults to True.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/auto-attendants/status"

        data = {
            "instances": [
                {"serviceUserId": auto_attendant_user_id, "isActive": status}
                for auto_attendant_user_id in auto_attendant_user_ids
            ]
        }

        return self._requester.put(endpoint, data=data)

    def put_auto_attendant(
        self,
        service_provider_id: str,
        group_id,
        auto_attendant_user_id: str,
        updates: dict,
    ):
        """Updates a single target Auto Attendant.

        Note: Needs the service instance profile to use this method.

        Args:
            service_provider_id (str): Service Provider ID where Group is hosted.
            group_id (str): Group ID where target Auto Attendant is located.
            auto_attendant_user_id (str): Target Auto Attendant User ID.
            updates (dict): Updates to be applied to Auto Attendant.

        Returns:
            Dict: Updated version of the Auto Attendant with updates applied.
        """

        endpoint = "/groups/auto-attendants"

        updates = format_sip(updates)

        updates["serviceProviderId"] = service_provider_id
        updates["groupId"] = group_id
        updates["serviceUserId"] = auto_attendant_user_id

        return self._requester.put(endpoint, data=updates)

    def put_auto_attendant_submenu(
        self, auto_attendant_user_id: str, submenu_id: str, updates: dict
    ):
        """This method allows you to update the configuration of the submenus for your AAs

        Args:
            auto_attendant_user_id (str): Service user ID of your auto attendant.
            submenu_id (str): Service user ID of your submenu
            updates (dict): Updates your applying to your submenu.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/auto-attendants/submenus"

        updates["serviceUserId"] = (auto_attendant_user_id,)
        updates["submenuId"] = submenu_id

        return self.request.put(endpoint, data=updates)

    # DELETE

    def delete_auto_attendant(self, service_user_id: str):
        """Removes an Auto Attendant (AA) from a group.

        Args:
            service_user_id(str): The service user ID of the AA.

        Returns:
            Dict: Returns the profile of the deleted AA.
        """

        endpoint = "/groups/auto-attendants"

        params = {"serviceUserId": service_user_id}

        return self._requester.delete(endpoint, params=params)

    def delete_auto_attendant_submenu(self, service_user_id: str, submenu_id: str):
        """Removes an Auto Attendant (AA) Submenu from the AA configuration. Submenus are only a feature of the 'Auto Attendant - Standard' service. These are not available on Basic AAs.

        Args:
            service_user_id(str): The service user ID of the AA.
            submenu_id (str): The ID of the Submenu to be removed from the AA.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/auto-attendants/submenus"

        params = {"serviceUserId": service_user_id, "submenuId": submenu_id}

        return self._requester.delete(endpoint, params=params)



from .base_endpoint import BaseEndpoint


class ServiceProviders(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_service_providers(self, reseller_id: str = None):
        """
        Fetches list of service providers.

        Args:
            reseller_id (str): Only list the Service Provider IDs within the specified Reseller.

        Returns:
            dict: A dictionary containing the response from the

        """
        endpoint = "/service-providers"
        params = {"resellerId": reseller_id}

        return self._requester.get(endpoint, params=params)

    def get_service_provider(self, service_provider_id: str):
        """
        Fetches a service provider by ID with further details.

        Args:
            reseller_id (str): Only list the Service Provider IDs within the specified Reseller.

        Returns:
            dict: Dictionary of details for the service provider.
        """

        endpoint = "/service-providers"

        params = {"serviceProviderId": service_provider_id}

        return self._requester.get(endpoint, params=params)


# POST

# PUT

# DELETE



from .base_endpoint import BaseEndpoint


class EmergencyZones(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    # POST
    def post_group_emergency_zones(
        self, service_provider_id: str, group_id: str, ip_addresses: list
    ):
        """Updates the IP address(es) for the Emergency Zone configured in the group.

        Args:
            service_provider_id (str): Service provider ID where the Emergency Zone to be updated exists.
            group_id (str): Group ID where the Emergency Zone to be updated exists.
            ip_addresses (list): A list of IP address ranges (dicts) to be added to the Emergency Zone. If the IP address to be applied is not a range, the min and max values should be the same.

        Returns:
            Dict: Emergency Zone profile with updated IP addresses.
        """

        endpoint = "/groups/emergency-zones"

        data = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "ipAddresses": ip_addresses,
        }

        return self._requester.post(endpoint, data=data)

    # PUT
    def put_group_emergency_zones(
        self,
        service_provider_id: str,
        group_id: str,
        is_active: bool = True,
        zone_rules: str = None,
        emergency_notification_email: str = None,
        ip_addresses: list = None,
    ):
        """Updates the Emergency Zone configuration in the group.

        Args:
            service_provider_id (str): Service provider ID where the Emergency Zone to be updated exists.
            group_id (str): Group ID where the Emergency Zone to be updated exists.
            is_active (bool, optional): Whether the Emergency Zone service is active or not. Defaults to True
            zone_rules (str, optional): The rules of the Emergency Zone. This will either be "Prohibit all registrations and call originations" or "Prohibit emergency call originations".
            emergency_notification_email (str, optional): The email address where emergency call notifications should be sent.
            ip_addresses (list, optional): A list of IP address ranges (dicts) to be added to the Emergency Zone. If the IP address to be applied is not a range, the min and max values should be the same.

        Returns:
            Dict: Updated Emergency Zone configuration.
        """

        endpoint = "/groups/emergency-zones"

        data = {"groupId": group_id, "serviceProviderId": service_provider_id}

        if is_active:
            data["isActive"] = is_active
        if zone_rules:
            data["emergencyZonesProhibition"] = zone_rules
        if emergency_notification_email:
            data["sendEmergencyCallNotifyEmail"] = True
            data["emergencyCallNotifyEmailAddress"] = emergency_notification_email
        if ip_addresses:
            data["ipAddresses"] = ip_addresses

        return self._requester.put(endpoint, data)

    # DELETE



import requests
import json

from ratelimit import limits, sleep_and_retry

from .exceptions import OSApiResponseError
from .utils.formatters import sanitise_data


class Requester:
    __instance = None  # Class variable to hold the singleton instance

    @staticmethod
    def get_instance(base_url=None, rate_limit=None, logger=None):
        if Requester.__instance is None:
            Requester(base_url, rate_limit, logger)
        return Requester.__instance

    def __init__(self, base_url, rate_limit, logger):
        """
        Initialize the Requester with default values.

        NOTE: This object is a singleton and can't be instantiated more than once.
        """
        if Requester.__instance is not None:
            raise Exception("Singleton cannot be instantiated more than once!")
        else:
            self.base_url = base_url
            self.rate_limit = rate_limit
            self.headers = {
                "Authorization": "",
                "Content-Type": "application/json",
            }
            self.logger = logger

            self.logger.info(
                f"Requester initialized with base_url: {self.base_url}, rate_limit: {self.rate_limit}"
            )

            Requester.__instance = self

    def get(self, endpoint, data=None, params=None):
        return self._request(requests.get, endpoint, data, params)

    def post(self, endpoint, data=None):
        return self._request(requests.post, endpoint, data)

    def put(self, endpoint, data=None):
        return self._request(requests.put, endpoint, data)

    def delete(self, endpoint, data=None, params=None):
        return self._request(requests.delete, endpoint, data, params)

    def _request(self, method, endpoint, data=None, params=None):
        """Handles an API request with or without rate limiting."""

        self.logger.info(
            f"Initiating API request, method: {method.__name__.upper()}, endpoint: {endpoint}"
        )

        if self.rate_limit:
            return self._rate_limited_request(method, endpoint, data, params)

        # Logging request details
        request_payload = json.dumps(data) if data is not None else None
        self.logger.debug(
            f"Sending request, method: {method.__name__.upper()},"
            f"endpoint: {self.base_url + endpoint}, params: {params}, data: {sanitise_data(data) if data else 'None'}"
        )

        response = method(
            url=self.base_url + endpoint,
            headers=self.headers,
            data=request_payload,
            params=params or {},
        )

        return self._handle_response(response, method.__name__, endpoint)

    @sleep_and_retry
    @limits(calls=5, period=1)
    def _rate_limited_request(self, method, endpoint, data=None, params=None):
        """Handles an API request with rate limiting."""

        self.logger.warning(
            f"Rate limit active. Request may be delayed, method: {method.__name__.upper()}, endpoint: {endpoint}"
        )

        request_payload = json.dumps(data) if data is not None else None
        self.logger.debug(
            f"Sending rate-limited request, method: {method.__name__.upper()},"
            f"endpoint: {self.base_url + endpoint}, params: {params}, data: {sanitise_data(data) if data else 'None'}"
        )

        response = method(
            url=self.base_url + endpoint,
            headers=self.headers,
            data=request_payload,
            params=params or {},
        )

        return self._handle_response(response, method.__name__, endpoint)

    def _handle_response(self, response, method_name, endpoint):
        """Handles response logging and error handling."""

        # Log response status
        if response.status_code >= 200 and response.status_code < 300:
            self.logger.info(
                f"API Call Success, method: {method_name.upper()}, endpoint: {endpoint}, status_code: {response.status_code}"
            )
            self.logger.debug(f"response_data: {sanitise_data(response.json())}")

            return response.json()

        # Log API errors
        else:
            self.logger.error(
                f"API Error, method: {method_name.upper()}, endpoint: {endpoint}, status_code: {response.status_code}, response_text: {response.text}"
            )
            raise OSApiResponseError(response)



from typing import Dict, Any, Optional

from . import scripts
from .api import API


class Scripter:
    """This object acts as the gateway to all pre-written scripts in /scripts/.


    Each api object created creates its own associated Scripter object on api creation.
    Additionally, this object can be created solely and passed an api, however, this
    will result in two exact Scripter objects.

    Intended use: odin_api.scripter.{script function}

    :param api: api object in package odin_api, this is used in the scripts to achieve objective.
    """

    __instance = None

    @staticmethod
    def get_instance(api: API = None) -> "Scripter":
        """
        Singleton implementation for Scripter object.

        Args:
            api (API): API object to be used in the scripts.

        Returns:
            Scripter: Scripter object.
        """
        if Scripter.__instance is None:
            Scripter(api)
        return Scripter.__instance

    def __init__(
        self,
        api: API,
    ) -> None:
        if Scripter.__instance is not None:
            raise Exception("Singleton cannot be instantiated more than once!")
        else:
            self.api = api
            Scripter.__instance = self

    def _run_script(self, script_name: str, *args, **kwargs) -> Dict[str, Any]:
        """Dynamically runs the specified script."""
        self.api.logger.debug(
            f"Script {script_name} executed, args: {[args]}, kwargs: {kwargs}"
        )
        try:
            script_function = getattr(scripts, script_name)
            self.api.logger.info(f"Script {script_name} executed")
        except AttributeError:
            self.api.logger.error(
                f"Script failed to execute, error: Script '{script_name}' not found."
            )
            raise AttributeError(
                f"Script '{script_name}' not found in 'scripts' module."
            )
        return script_function(self.api, *args, **kwargs)

    def bulk_password_reset(
        self,
        *,
        service_provider_id: str,
        group_id: str,
        users: list,
        password_type: str,
    ) -> Dict[str, any]:
        """Resets a list of users SIP passwords or Voicemail passcodes. Specify in password_type with the options of
        'SIP' or 'Voicemail' and the script will perform the necessary actions.

        Args:
            service_provider_id (str): Service Provider ID where group is hosted.
            group_id (str): Group ID where target users are located.
            users (list): List of User IDs of the target users to reset the password.
            password_type (str): Type of password to reset either 'SIP' or 'Voicemail'. Only accepts these two options.

        Raises:
            OSInvalidPasswordType: Only valid password options are SIP, VM, WEB. If another is requested this will be raised.

        Returns:
            Dict: Users and their new passwords.
        """

        return self._run_script(
            "bulk_password_reset", service_provider_id, group_id, users, password_type
        )

    def find_alias(
        self, *, service_provider_id: str, group_id: str, alias: str
    ) -> Dict[str, any]:
        """Locates alias if assigned to broadworks entity.

        Args:
            service_provider_id (str): Service Prodiver where group is hosted.
            group_id (str): Group where alias is located.
            alias (int): Alias number to identify e.g. 0

        Raises:
             AOALiasNotFound: If alias not found AOAliasNotFound error raised

        Returns:
            Dict: Returns type and name/ userId of entity where alias located.

        """
        return self._run_script("find_alias", service_provider_id, group_id, alias)

    def group_audit(self, *, service_provider_id: str, group_id: str) -> Dict[str, any]:
        """
        Produces a report of key information within the group.
        Reports on DN usage, Service and Service pack usage, Trunking call capacity and group info.

        Args:
            service_provider_id (str): Service Provider ID or Enterprise ID containing the Group ID.
            group_id (str): Group ID to generate the report for.

        Returns:
            Dict: Formatted report of the group.
        """
        return self._run_script("group_audit", service_provider_id, group_id)

    def locate_free_extension(
        self,
        *,
        service_provider_id: str,
        group_id: str,
        range_start: int,
        range_end: int,
    ) -> Dict[str, any]:
        """Locates the lowest value free extension given the provided range of extension numbers.

        Raises: OSExtensionNotFound: Raises when a free extension is not located within the passed range.

        Args:
            service_provider_id (str): Service Provider/ Enterprise ID where Group is located which hosts needed free extensions
            group_id (str): Group ID where target extensions are located.
            range_start (int): integral value specifying the starting range for free extensions
            range_end (int): integral value specifying the ending range for free extensions

        Returns:
            Dict: Data of the free extension {extension: "1000"}
        """
        return self._run_script(
            "locate_free_extension",
            service_provider_id,
            group_id,
            range_start,
            range_end,
        )

    def move_numbers(
        self,
        *,
        current_service_provider_id: str,
        current_group_id: str,
        target_service_provider_id: str,
        target_group_id: str,
        start_of_range_number: str,
        end_of_range_number: Optional[str] = None,
    ) -> bool:
        """Moves a list of numbers from existing group to another group on the same broadworks instance.
        This can move numbers between Service Provider/ Enterprise and groups in the same Service Provider/ Enterprise.

        Note: Numbers need to be strings and follow this format: +{country code}-{number}.

        Args:
            current_service_provider_id (str): Current Service Provider/ Enterprise where numbers are located.
            current_group_id (str): Current Group ID where numbers are located.
            target_service_provider_id (str): Target Service Provider/ Enterprise to move the numbers to.
            target_group_id (str): Target Group to move numbers to.
            start_of_number_range (str): Starting number in range of numbers you would like to move.
            end_of_number_range (str): Ending nummber in range of numbers you would like to move. If you need to move
            only one number do not enter a value for this paramter. Defaults to None.

        Returns:
            Bool: Returns a True once complete.
        """
        return self._run_script(
            "move_numbers",
            current_service_provider_id,
            current_group_id,
            target_service_provider_id,
            target_group_id,
            start_of_range_number,
            end_of_range_number,
        )

    def remove_numbers(
        self,
        *,
        service_provider_id: str,
        group_id: str,
        start_of_range_number: str,
        end_of_range_number: Optional[str] = None,
    ) -> bool:
        """Removes a singular or range of numbers from the entire Broadworks instance.

        Note: Numbers need to be strings and follow this format: +{country code}-{number}.

        Args:
            service_provider_id (str): Service Provider/ Enterprise ID where Group is located which hosts target numbers.
            group_id (str): Group ID where target numbers are located.
            start_of_number_range (str): Starting number in range of numbers you would like to remove.
            end_of_number_range (str): Ending nummber in range of numbers you would like to remove. If you need to remove
            only one number do not enter a value for this paramter. Defaults to None.

        Returns:
            Bool: Returns a True once complete.
        """
        return self._run_script(
            "remove_numbers",
            service_provider_id,
            group_id,
            start_of_range_number,
            end_of_range_number,
        )

    def service_provider_trunking_capacity(
        self, *, service_provider_id: str
    ) -> Dict[str, any]:
        """Returns a JSON breakdown of the Trunking Call Capacity of a Service Provider/ Enterprise (SP/ENT). 
        This will show the totals at each level from SP/ ENT to Group to Trunk Groups located in Groups. 
        At each level Max Active Calls and Bursting Max Active calls are detailed and then differences at 
        calculated.
        
        Raises: 
            OSServiceNotAssigned: Raises when the SP/ Ent does not have the service 'Trunk Group' assigned. 

        Args:
            service_provider_id (str): Target Service Provider ID/ Enterprise ID that you would like the \
                Trunk Call Capacity breakdown.

        Returns:
            Dict: Data of Trunking Call Capacity details of SP/ ENT, Groups, and Trunk Groups.
        """
        return self._run_script(
            "service_provider_trunking_capacity", service_provider_id
        )

    def user_association(
        self, *, service_provider_id: str, group_id: str, user_id: str
    ) -> Dict[str, any]:
        """
        Identify a user's associations with Call Centers (CC), Hunt Groups (HG),
        and Pick Up Groups.

        Args:
            service_provider_id (str): Service Provider where the group is hosted.
            group_id (str): Group where the User is located.
            user_id (str): Target user ID.

        Returns:
            Dict: Formatted output of the user showing all CC, HG, and Pick Up user is assigned to.
        """
        self._run_script("user_association", service_provider_id, group_id, user_id)

    def user_registration(
        self, *, service_provider_id: str, group_id: str
    ) -> Dict[str, any]:
        """Generates a dictionary detailing each Users ID, device name and registration status within a group.

        Args:
            service_provider_id (str):  Service Provider/ Enterprise where group is hosted.
            group_id (str): Target Group you would like to know user statistics for.

        Returns:
            Dict: User's ID, Device Name, and Registration status.
        """
        return self._run_script("user_registration", service_provider_id, group_id)

    def webex_builder(
        self,
        *,
        service_provider_id: str,
        group_id: str,
        user_id: str,
        device_type: str,
        email: str,
        primary_device: bool = True,
        webex_feature_pack_name: Optional[str] = None,
        enable_integrated_imp: bool = True,
    ) -> Dict[str, any]:
        """Builds a Webex device and assigns to user either as the primary device or a
        Shared Call Appearance.

        Args:
            service_provider_id (str): Service Provider ID where group is hosted.
            group_id (str): Group ID where target user is hosted.
            user_id (str): Target user to build and add webex device.
            device_type (str): Name of the device profile to apply.
            email (str): Email of user. This will be used to sign into the webex client.
            primary_device (bool, optional): Setting to False will assign as SCA, True is primary. Defaults to True.
            webex_feature_pack_name (str, optional): Feature pack to assign for Webex services. Defaults to None.
            enable_integrated_imp (bool, optional): Enables Integrated IMP service for the user if True .Defaults to True.

        Returns:
            Dict: Webex user/ device details. Includes webex client username and password, and if primary device
            device type set.

        Raises:
            OSLicenseNonExistent: Raised if user does not have correct license which allows for SCA devices.
        """

        return self._run_script(
            "webex_builder",
            service_provider_id,
            group_id,
            user_id,
            device_type,
            email,
            primary_device,
            webex_feature_pack_name,
            enable_integrated_imp,
        )



from .administrators import Administrators
from .alternate_numbers import AlternateNumbers
from .announcements import Announcements
from .authentication import Authentication
from .auto_attendant import AutoAttendants
from .call_centers import CallCenters
from .call_forwarding_always import CallForwardingAlways
from .call_processing_policies import CallProcessingPolicies
from .call_pickup import CallPickup
from .call_fowarding_selective import CallForwardingSelective
from .call_forwarding_not_reachable import CallForwardingNotReachable
from .call_forwarding_no_answer import CallForwardingNoAnswer
from .call_forwarding_busy import CallForwardingBusy
from .call_records import CallRecords
from .devices import Devices
from .dns import DNs
from .groups import Groups
from .emergency_zones import EmergencyZones
from .do_not_disturb import DoNotDisturb
from .hunt_groups import HuntGroups
from .service_providers import ServiceProviders
from .services import Services
from .session import Session
from .shared_call_appearance import SharedCallAppearance
from .schedules import Schedules
from .reports import Reports
from .registration import Registration
from .password_generate import PasswordGenerate
from .trunk_groups import TrunkGroups
from .users import Users

__all__ = [
    "Administrators",
    "AlternateNumbers",
    "Announcements",
    "Authentication",
    "AutoAttendants",
    "CallCenters",
    "CallForwardingAlways",
    "CallProcessingPolicies",
    "CallPickup",
    "CallForwardingSelective",
    "CallForwardingNotReachable",
    "CallForwardingNoAnswer",
    "CallForwardingBusy",
    "CallRecords",
    "Devices",
    "DNs",
    "Groups",
    "EmergencyZones",
    "DoNotDisturb",
    "HuntGroups",
    "ServiceProviders",
    "Services",
    "Session",
    "SharedCallAppearance",
    "Schedules",
    "Reports",
    "Registration",
    "PasswordGenerate",
    "TrunkGroups",
    "Users",
]


# import importlib
# import pkgutil

# # Get the package name dynamically
# package_name = __name__

# # Iterate over all modules in the `endpoints` directory
# for _, module_name, _ in pkgutil.iter_modules(__path__):
#     if module_name != "base_endpoint":  # Exclude base_endpoint.py
#         importlib.import_module(f"{package_name}.{module_name}")
